(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

(** Correctness of CPS (src to mid) translation. *)

Require Import Arith Bool List Relations Program. 
Require Import vgtac basis.
Require Import lang_src_lsi lang_mid lang_spec.
Require Import model model_common model_properties gw.
Require Import ms cps ssr.

Set Implicit Arguments. 
Global Set Bullet Behavior "Strict Subproofs".



(** * CPS preserves well-formedness. *)

Inductive renG:
  list termvar -> (termvar -> itermvar) -> list itermvar -> Prop :=
| sG_nil: forall ren, renG nil ren nil
| sG_cons: forall x y xs ys ren (SG: renG xs ren ys),
             renG (x::xs) [y//x]>>ren (y::ys)
.
Hint Constructors renG.

Lemma renG_In: forall x (ty: typ) G ren G'
    (IN: list_assoc eq_nat_dec x G = Some ty)
    (SG: renG (map fst G) ren G'),
  In (ren x) G'.
Proof.
  i; ginduction G; vauto.
  i; destruct a; ss; des_ifH IN; clarify; inv SG
  ; [by rewrite tmext_eq; vauto|].
  right; replace (tmext n y _ _) with (ren0 x)
  ; [|by rewrite tmext_neq; auto].
  eauto.
Qed.

Lemma cps_wfexp: forall F G e ty m k ren X G'
    (HSP: hastypepre F G e ty)
    (SG: renG (map fst G) ren G')
    (KWF: forall m x,
            wfexp ((inl x::X) ++ map inl G' ++ map (inl<*>inl<*>fst) F) (snd (k m x))),
  wfexp (X ++ map inl G' ++ map (inl<*>inl<*>fst) F) (snd (cps e m ren k)).
Proof.
  i; ginduction HSP; i; ss.
  - destruct (k _ _) eqn:E; ss; clarify; ss.
    specialize (KWF (next m) (inr m)); rewrite E in KWF; eauto.
  - eapply IHHSP; eauto; inst; i; s.
    destruct (k _ _) eqn:E; ss; clarify; ss; splits; auto.
    specialize (KWF (next m0) (inr m0)); rewrite E in KWF.
    rewrite cons_app; apply wfexp_weaken; rewrite <-cons_app; eauto.
  - specialize (KWF m (ren x)).
    exploit renG_In; eauto; i.
    apply in_split in x1; des; subst.
    rewrite map_app in KWF *; ss.
    rewrite <-!app_assoc, <-app_comm_cons, app_assoc.
    rewrite <-!app_assoc, <-app_comm_cons, app_assoc in KWF.
    match goal with |- wfexp ?l _ => change l with (nil ++ l) end.
    apply wfexp_contract; auto.
  - specialize (KWF m (inl l)).
    apply list_assoc_someD, in_split in H; des; subst.
    rewrite ->map_app in *; ss; rewrite 2!app_assoc.
    match goal with |- wfexp ?l _ => change l with (nil ++ l) end.
    apply wfexp_contract; rewrite <-!app_assoc; auto.
  - destruct (k _ _) eqn:E; ss; clarify; ss; splits; auto.
    specialize (KWF (next m) (inr m)); rewrite E in KWF; eauto.
  - destruct (k _ _) eqn:E; ss; clarify; ss; splits; auto.
    specialize (KWF (next m) (inr m)); rewrite E in KWF; eauto.
  - eapply IHHSP1; eauto; inst; i; s.
    rewrite app_comm_cons; eapply IHHSP2; eauto; inst; i; s.
    destruct (k _ _) eqn:E'; ss; clarify; ss.
    splits; auto.
    specialize (KWF (next m1) (inr m1)); rewrite E' in KWF.
    rewrite cons_app; do 2 apply wfexp_weaken; rewrite <-cons_app; eauto.
  - eapply IHHSP1; eauto; inst; i; s.
    destruct (cps ea _ _ _) eqn:Ea; ss.
    destruct (cps eb _ _ _) eqn:Eb; ss.
    splits; auto.
    + fold (snd (n, e0)); rewrite <-Ea.
      rewrite app_comm_cons; eapply IHHSP2; eauto; inst; i; s.
      specialize (KWF m1 x0).
      rewrite cons_app; apply wfexp_weaken; rewrite <-cons_app; eauto.
    + fold (snd (n0, e1)); rewrite <-Eb.
      rewrite app_comm_cons; eapply IHHSP3; eauto; inst; i; s.
      specialize (KWF m1 x0).
      rewrite cons_app; apply wfexp_weaken; rewrite <-cons_app; eauto.
  - eapply IHHSP1; eauto; inst; i; s.
    rewrite app_comm_cons; eapply IHHSP2; eauto; inst; i; s.
    destruct (k _ _) eqn:E'; ss; clarify; ss.
    splits; auto.
    specialize (KWF (next m1) (inr m1)); rewrite E' in KWF.
    rewrite cons_app; do 2 apply wfexp_weaken; rewrite <-cons_app; eauto.
  - eapply IHHSP; eauto; inst; i; s.
    destruct (k _ _) eqn:E; ss; clarify; ss; split; [by left; econs|].
    specialize (KWF (next m0) (inr m0)); rewrite E in KWF.
    rewrite cons_app; apply wfexp_weaken; rewrite <-cons_app; eauto.
  - eapply IHHSP; eauto; inst; i; s.
    destruct (k _ _) eqn:E; ss; clarify; ss; split; [by left; econs|].
    specialize (KWF (next m0) (inr m0)); rewrite E in KWF.
    rewrite cons_app; apply wfexp_weaken; rewrite <-cons_app; eauto.
  - eapply IHHSP; eauto; inst; i; s.
    destruct (k _ _) eqn:E; ss; clarify; ss; splits; auto.
    specialize (KWF (next m0) (inr m0)); rewrite E in KWF.
    rewrite cons_app; apply wfexp_weaken; rewrite <-cons_app; eauto.
  - eapply IHHSP; eauto; inst; i; s.
    destruct (k _ _) eqn:E; ss; clarify; ss; splits; auto.
    specialize (KWF (next m0) (inr m0)); rewrite E in KWF.
    rewrite cons_app; apply wfexp_weaken; rewrite <-cons_app; eauto.
  - eapply IHHSP1; eauto; inst; i; s.
    destruct (cps ea _ _ _) eqn:Ea; ss.
    destruct (cps eb _ _ _) eqn:Eb; ss.
    splits; auto.
    + fold (snd (n, e0)); rewrite <-Ea.
      cut (wfexp (inl x0 :: X ++ (inl (inr m0) :: map inl G') ++ map (inl<*>inl<*>fst) F)
                 (snd (cps ea (next m0) (tmext x (inr m0) ren) k))).
      * i; eapply wfexp_exchange; eauto; symmetry.
        eapply perm_trans; [by eapply perm_swap|].
        eapply perm_skip, Permutation_middle.
      * rewrite ->app_comm_cons, <-map_cons.
        eapply IHHSP2; eauto; inst; i; s.
        specialize (KWF m1 x1).
        rewrite cons_app; apply wfexp_weaken.
        rewrite app_assoc; apply wfexp_weaken; rewrite <-cons_app; eauto.
    + fold (snd (n0, e1)); rewrite <-Eb.
      cut (wfexp (inl x0 :: X ++ (inl (inr m0) :: map inl G') ++ map (inl<*>inl<*>fst) F)
                 (snd (cps eb n (tmext x (inr m0) ren) k))).
      * i; eapply wfexp_exchange; eauto; symmetry.
        eapply perm_trans; [by eapply perm_swap|].
        eapply perm_skip, Permutation_middle.
      * rewrite ->app_comm_cons, <-map_cons.
        eapply IHHSP3; eauto; inst; i; s.
        specialize (KWF m1 x1).
        rewrite cons_app; apply wfexp_weaken.
        rewrite app_assoc; apply wfexp_weaken; rewrite <-cons_app; eauto.
  - eapply IHHSP; eauto; inst; i; s.
  - eapply IHHSP; eauto; inst; i; s.
  - destruct (cps _ _ _ _) eqn:E; ss.
    destruct (k _ _) eqn:E'; ss; clarify; ss; split.
    + fold (snd (n, e0)); rewrite <-E.
      cut (wfexp ((inr (next (next (next m))) :: purge X) ++
                  (map inl (inr (next (next m)) :: inr (next m) :: G'))
                          ++ map (inl<*>inl<*>fst) F)
                 (snd (cps e (next (next (next (next m))))
                           [inr (next (next m))//x]>>[inr (next m)//f]>>ren
                           (fun m' y => (m', IEret (next (next (next m))) y))))).
      * i; eapply wfexp_exchange; eauto; symmetry.
        s; apply perm_skip, Permutation_cons_app.
        rewrite ->2!purge_app, !purge_map_inl.
        eapply Permutation_cons_app.
        rewrite ->2!map_comp, purge_map_inl, <-2!map_comp; auto.
      * eapply IHHSP; eauto; inst; i; split; s; auto.
    + specialize (KWF n (inr m)); rewrite E' in KWF; eauto.
  - eapply IHHSP1; eauto; inst; i; s.
    rewrite app_comm_cons; eapply IHHSP2; eauto; inst; i; s.
    destruct (k _ _) eqn:E'; ss; clarify; ss; splits; auto.
    specialize (KWF (next (next m1)) (inr (next m1))); rewrite E' in KWF.
    rewrite cons_app; do 2 apply wfexp_weaken; rewrite <-cons_app; eauto.
  - destruct (cps _ _ _ _) eqn:E; ss.
    destruct (k _ _) eqn:E'; ss; clarify; ss; split.
    + fold (snd (n, e0)); rewrite <-E.
      rewrite ->!purge_app, 2!map_comp, !purge_map_inl, <-2!map_comp, app_comm_cons.
      eapply IHHSP; eauto; inst; i; split; s; auto.
    + specialize (KWF n (inr m)); rewrite E' in KWF; eauto.
  - eapply IHHSP; eauto; inst; i; s.
    destruct (k _ _) eqn:E'; ss; clarify; ss; split; auto.
    fold (snd (n, e0)); rewrite <-E'.
    specialize (KWF (next (next m0)) (inr (next m0))).
    rewrite cons_app; apply wfexp_weaken; rewrite <-cons_app; eauto.
  - eapply IHHSP; eauto; inst; i; s.
  - eapply IHHSP1; eauto; inst; i; s.
    eapply wfexp_exchange.
    + eapply IHHSP2; eauto; inst; i; s.
      specialize (KWF m1 x1).
      rewrite ->cons_app, app_assoc; apply wfexp_weaken; rewrite <-cons_app; eauto.
    + s; eapply Permutation_sym.
      eapply perm_trans; [by apply Permutation_cons_app|]; auto.
  - eapply IHHSP; eauto; inst; i; s.
    destruct (k _ _) eqn:E'; ss; clarify; ss; split; auto.
    fold (snd (n, e0)); rewrite <-E'.
    specialize (KWF (next m0) (inr m0)).
    rewrite cons_app; apply wfexp_weaken; rewrite <-cons_app; eauto.
  - eapply IHHSP; eauto; inst; i; s.
    destruct (k _ _) eqn:E'; ss; clarify; ss; split; auto.
    fold (snd (n, e0)); rewrite <-E'.
    specialize (KWF (next m0) (inr m0)).
    rewrite cons_app; apply wfexp_weaken; rewrite <-cons_app; eauto.
  - eapply IHHSP1; eauto; inst; i; s.
    rewrite app_comm_cons; eapply IHHSP2; eauto; inst; i; s.
    destruct (k _ _) eqn:E'; ss; clarify; ss.
    splits; auto.
    specialize (KWF (next m1) (inr m1)); rewrite E' in KWF.
    rewrite cons_app; do 2 apply wfexp_weaken; rewrite <-cons_app; eauto.
  - eapply IHHSP1; eauto; inst; i; s.
    rewrite app_comm_cons; eapply IHHSP2; eauto; inst; i; s.
    destruct (k _ _) eqn:E'; ss; clarify; ss.
    split; auto.
    specialize (KWF (next m1) (inr m1)); rewrite E' in KWF.
    rewrite cons_app; do 2 apply wfexp_weaken; rewrite <-cons_app; eauto.
Qed.

Lemma cps_good: forall e n ren k n'
    (CNT: forall n x n', snd (k n x) <> IEnat n'),
  snd (cps e n ren k) <> IEnat n'.
Proof.
  induction e; vauto; eauto; s; i
  ; try (by destruct (k _ _); vauto)
  ; try (by destruct (cps _ _ _ _) eqn:E
         ; try (rewrite <-E; eapply IHe; i)
         ; destruct (k _ _); vauto)
  ; try (by destruct (cps _ _ _ _) eqn:E1
         ; rewrite <-E1; eapply IHe1; i
         ; destruct (cps e2 _ _ _) eqn:E2
         ; rewrite <-E2; eapply IHe2; i
         ; destruct (k _ _); vauto)
  ; try (by destruct (cps _ _ _ _) eqn:E1
         ; rewrite <-E1; eapply IHe1; i
         ; destruct (cps e2 _ _ _)
         ; destruct (cps e3 _ _ _)).
Qed.

Lemma tlcps_good: forall n (v: lang_src.val) (fty: flextyp) FI
    (WT: hastype FI [] v fty),
  is_good_thunk (snd (tlcps n v)).
Proof.
  unfold tlcps; induction v; i; clarify; inv WT; inversion HT
  ; try (by assert (X:=typ_is_flex fty); destruct (flext_of fty); vauto)
  ; ss; clarify; destruct (cps _ _ _ _) eqn:E; clarify.
  - ii; inv FH.
  - ii; inv FH; assert (X:= eq_refl (snd (n0, IEnat n1))).
    rewrite <-E in X at 1; eapply cps_good in X; s; vauto.
Qed.

Lemma tlcps_wfatom: forall n (v: lang_src.val) (fty: flextyp) FI
    (WT: hastype FI [] v fty),
  wfatom (map (inl<*>inl<*>fst) FI) (snd (tlcps n v)).
Proof.
  unfold tlcps; destruct v; i; clarify; inv WT; inversion HT
  ; try (by destruct e_of; ss)
  ; ss; clarify; destruct (cps _ _ _ _) eqn:E; clarify
  ; s; fold (snd (n0, e0))
  ; rewrite <-E, cons_app, 2!map_comp, purge_map_inl, <-2!map_comp.
  - rewrite ->(cons_app (inl (inr n))), app_comm_cons.
    match goal with |- context[[@inl ?X ?Y ?a; @inl ?X ?Y ?b]] => change ([@inl X Y a; @inl X Y b]) with (map (@inl X Y) [a; b]) end.
    eapply cps_wfexp; eauto; inst; s; eauto.
  - match goal with |- context[map ?f ?l] => change (map ?f ?l) with (map inl nil ++ map f l) end.
    eapply cps_wfexp; eauto; inst; s; eauto.
Qed.

Theorem mcps_wfmod: forall FI FO FI' FO' m
    (EFI: FI' = map_pair id flext_of FI)
    (EFO: FO' = map_pair id flext_of FO)
    (WF: mhastype FI' m FO'),
  mid_wfmod (map fst FI) (mcps m) (map fst FO).
Proof.
  unfold mcps; generalize 0.
  i; ginduction m; i; inv WF; inv MHT; destruct FO; clarify.
  destruct p; unfold mcps; ss; des; clarify.
  destruct (tlcps n v) eqn:E.
  hexploit (tlcps_good n v); eauto; i.
  hexploit (tlcps_wfatom n v); eauto; i.
  rewrite ->E in *; ss; econs; eauto.
  - rewrite <-map_map, <-map_comp.
    rewrite ->map_comp, map_fst_map_pair, <-map_comp in X0; ss.
  - replace (map fst FI) with
    (map fst (map (fun p => (fst p, flext_of (snd p))) FI)).
    eapply list_assoc_noneD; eauto.
    clear; induction FI; s; try rewrite IHFI; auto.
  - replace (_ ++ [l0]) with (map fst (FI ++ [(l0, f)]))
    ; [|by rewrite map_app; auto].
    eapply IHm; eauto; econs; auto.
    + rewrite (map_pair_app eqdec); s; eauto.
    + rewrite (map_pair_app eqdec).
      apply is_thunkenv_app; s; auto.
Qed.



(** * CPS establishes unique variable condition. *)

Definition varid (x: var) : nat :=
  match x with
    | inl (inl n)
    | inl (inr n)
    | inr n
      => n
  end.

Lemma cps_uniqvar: forall e n ren k n' e'
    (E: cps e n ren k = (n', e'))
    (K: forall m x m' e''
            (E: k m x = (m', e'')),
           uniq (map varid (bvars e'')) /\
           within m m' (map varid (bvars e''))),
  uniq (map varid (bvars e')) /\
  within n n' (map varid (bvars e')).
Proof.
  induction e; i
  ;  try (by ss; unfold next in *
         ; try (eapply IHe; eauto; i; inst; ss)
         ; destruct (k _ _) eqn:Ek; ss; clarify
         ; hexploit K; eauto; i; des; split
         ; [|by unfold bvars; fold bvars; rewrite !map_app; simpl map
             ; eauto using uniq_app_within, within_app]
         ; s; try rewrite app_nil_r; splits; i; des; auto
         ; rrdes X0; hexploit X1; eauto using in_map; hexploit X2; eauto using in_map
         ; s; i; omega)
  ; try (by ss; unfold next in *
         ; eapply IHe1; eauto; inst; fold cps; i; ss 
         ; eapply IHe2; eauto; i; inst; ss
         ; destruct (k _ _) eqn:Ek; ss; clarify
         ; hexploit K; eauto; i; des; split
         ; [|by unfold bvars; fold bvars; rewrite !map_app; simpl map
             ; eauto using uniq_app_within, within_app]
         ; s; try rewrite app_nil_r; splits; i; des; auto
         ; rrdes X0; hexploit X1; eauto using in_map; hexploit X2; eauto using in_map
         ; s; i; omega)
  ; try (by ss; unfold next in *
         ; eapply IHe1; eauto; inst; i; ss
         ; destruct (cps e2 _ _ _) eqn:Ee2
         ; exploit IHe2; eauto; i; des
         ; destruct (cps e3 _ _ _) eqn:Ee3; ss; clarify
         ; exploit IHe3; eauto; i; des; split
         ; [|by unfold bvars; fold bvars; rewrite !map_app; simpl map
             ; eauto using uniq_app_within, within_app]
         ; s; rewrite map_app; eapply uniq_app_within; eauto using in_map)
  ; try (by ss; unfold next in *
         ; destruct (cps e _ _ _) eqn:Ee
         ; destruct (k _ _) eqn:Ek; ss; clarify
         ; hexploit K; eauto; i
         ; exploit IHe; eauto; [by i; inst; ss; vauto|]; i; des
         ; split; unfold bvars; fold bvars; rewrite !map_app; simpl map
         ; eauto 20 using uniq_app_within, within_app)
  ; try (by eapply IHe; eauto; i; inst; ss)
  ; try (by hexploit K; eauto; i; des)
  ; try (by ss; unfold next in *; vauto).
  - ss; unfold next in *.
    eapply IHe1; eauto; i; inst; ss.
    destruct (cps e2 _ _ _) eqn:Ee2.
    exploit IHe2; eauto; i; des.
    destruct (cps e3 _ _ _) eqn:Ee3; ss; clarify.
    exploit IHe3; eauto; i; des.
    split; [|by unfold bvars; fold bvars; rewrite !map_app; simpl map
             ; eauto using uniq_app_within, within_app].
    s; rewrite map_app; split; [|by eapply uniq_app_within; eauto].
    i; apply in_app_or in FH; rrdes x1; des.
    + hexploit x5; eauto; hexploit x6; eauto; i; omega.
    + rrdes x3; hexploit x7; eauto; hexploit x8; eauto; i; omega.
  - ss; unfold next in *
    ; eapply IHe1; eauto; inst; fold cps; i; ss 
    ; eapply IHe2; eauto; i; inst; ss
    ; destruct (k _ _) eqn:Ek; ss; clarify
    ; hexploit K; eauto; i; des.
Grab Existential Variables.
  exact 0. exact 0. exact 0. exact 0. exact 0. exact 0.
Qed.

Lemma tlcps_uniqvar: forall n v t e
    (E: tlcps n v = (t, e)),
  uniq (map varid (bvars e)) /\
  within n t (map varid (bvars e)).
Proof.
  destruct v; s; i; clarify
  ; (destruct (cps _ _ _ _) eqn:EE; clarify; unfold next in *
     ; hexploit cps_uniqvar; eauto; inst; [by s; i; clarify|]
     ; intro UNQ; des; split; unfold bvars; fold bvars; rewrite map_app
     ; [|by eapply within_app; s; eauto using uniq_app_within]
     ; s; splits; i; des; auto; try omega
     ; by (rrdes UNQ0; exploit UNQ1; eauto; exploit UNQ2; eauto; i; omega)).
Qed.

Lemma mcps_uniqvar_aux: forall FI FO n m
    (WF: mhastype (map_pair id flext_of FI) m
                  (map_pair id flext_of FO)),
  uniq (map varid (bvars_mod (mcps_ n m))) /\
  forall x (IN: In x (map varid (bvars_mod (mcps_ n m)))), n <= x.
Proof.
  i; ginduction m; i; inv WF; inv MHT.
  destruct FO; vauto; destruct p; unfold mcps; ss; des; clarify.
  destruct (tlcps n v) eqn:E.
  hexploit tlcps_uniqvar; eauto; i; des.
  exploit (IHm (FI ++ [(l0, f)]) FO)
  ; [by rewrite (map_pair_app eqdec); s; econs; eauto
     ; apply is_thunkenv_app; s; auto|].
  i; des; esplits.
  - s; rewrite map_app; eapply uniq_app_le; eauto.
    i; eapply X0; eauto.
  - rrdes X0; i; ss; rewrite map_app in IN.
    apply in_app_or in IN; des; auto; setoid_rewrite X0; auto.
Qed.

Lemma mcps_uniqvar: forall FI FO m
    (WF: mhastype (map_pair id flext_of FI) m
                  (map_pair id flext_of FO)),
   uniqvar (mcps m).
Proof.
  i; eapply (uniq_mapD _ varid).  
  eapply mcps_uniqvar_aux; eauto.
Qed.



(** * Output of CPS translation simulates the corresponding input. *)



Notation cpsw := (wlift (lwemp ms_gw)).

Lemma cpsw_ext: forall s s'
    (EXT: ext cpsw s' s),
  snd (fst (fst s')) = snd (fst (fst s)).
Proof.
  i; rrdes EXT EXT EXT EXT2; auto.
Qed.


(** Continuation relation for correctness proof. *)

Definition ms_ksim (w: world mid_lsi src_lsi) i kia kib g si s tyv ty
    (ka: env * (nat -> itermvar -> nat * exp))
    (kb: src_lsi.(cont)) :=
  forall g' (LE: g <g= g')
         s' (EXTP: w.(extpub) s' si) (EXT: w.(ext) s' s)
         va vb (RV: vclos w s' (g' s') tyv va vb)
         x enva (LEx: lookup (inl x) enva = Some va)
         n ea (Ea: ea = (pmemp, Just (enva, snd (snd ka n x))))
         eb (Eb: eb = (pmemp, Nothing, Just (app_cont kb vb)))
         (LE: forall y v (IN: lookup y (fst ka) = Some v),
                lookup y enva = Some v)
         (UNQ: uniq (bvars (snd (snd ka n x))))
         (DJ: forall y v (LU: lookup y enva = Some v)
                         (IN: In y (bvars (snd (snd ka n x)))), False),
    esim w i kia kib g' si s' true ty ea eb.

Definition cps_ksim := ms_ksim cpsw.
Lemma cps_ksim_done: forall i ka kb g si s ty xk enva
    (LU: lookup (inr xk) enva = Some ka),
  cps_ksim i ka kb g si s ty ty
           (enva, fun mx y => (mx, IEret xk y)) kb.
Proof.
  ii; subst; pfold; econs; ii; esplits; eauto.
  eapply esim_ret; [by apply rt_refl| | |by eauto| | |]; eauto; rr; s; eauto.
Qed.


(** Expression relation for correctness proof *)

Notation proj_env s := (snd (fst (fst s))).

Definition ms_psim (lw: world_loca ms_gw) F G ty ea eb : Prop := exists i',
  forall tsub (n: nat) ren g s si enva denvb i kia kib ty' ka kb
    (GK: gknow (wlift lw) g)
    (EXT: extpub (wlift lw) s si)
    
    (RSV: forall l ty (IN: lookup l F = Some ty), exists va vb,
           <<INa: lookup (inl (inl l)) enva = Some va>> /\
           <<INb: lookup l (proj_env s) = Some vb>> /\
           <<SIM: vclos (wlift lw) s (g s) (tsubst_name tsub ty) va vb>>)
    
    (RDV: forall x ty (IN: lookup x G = Some ty), exists va,
           <<INa: lookup (inl (ren x)) enva = Some va>> /\
                  
           <<SIM: vclos (wlift lw) s (g s) (tsubst_name tsub ty) va (denvb x)>>)
    (UNQ: uniq (bvars (ea n ren ka)))
    (DJ: forall y v (LU: lookup y enva = Some v) (IN: In y (bvars (ea n ren ka))), False)
    (WCL: subst_weakly_closed denvb)
    (CL: forall x (IN: In x (map fst G)), forall y, ~ fv_exp (denvb x) y)
    
    (KSIM: ms_ksim (wlift lw) (wfo_nat i) kia kib g si s (tsubst_name tsub ty) ty' (enva, ka) kb)
    (LCT: lc_tsub tsub),
  esim (wlift lw) (wfo_nat (i' + i)) kia kib g si s true ty'
        (pmemp, Just (enva, ea n ren ka))
        (pmemp, Nothing, Just (app_cont kb (subst_exp denvb eb))).


(** ** Correctness proof for each expression form *)


Notation cpslw := (lwemp ms_gw).
Definition cps_correct F G e ty :=
  ms_psim cpslw F G ty (fun x ren k => snd (cps e x ren k)) e.


Lemma cps_correct_var: forall F G x ty
    (LU: lookup x G = Some ty),
  cps_correct F G (Evar x) ty.
Proof.
  eexists; ii; s; hexploit RDV; eauto; i; des.
  exploit KSIM; last (by instantiate (1:=0); eauto); eauto.
Qed.

Lemma cps_correct_lbl: forall F G f ty
    (LU: lookup f F = Some ty),
  cps_correct F G (Elbl f) ty.
Proof.
  eexists; ii; s; hexploit RSV; eauto; i; des.
  
  prepare; eapply pesim_stepb_obsolete; i; inst_pairs.
  - inv XTR; rrdes RL; s in Ec; noemp; clarify; esplits.
    + econs; inst_pairs; econs; eauto.
    + rr; esplits; eauto; s; noemp; eauto.
  - rewrite ->mid_csplit, src_csplit; apply pesim_release; auto.
    + eapply KSIM; i; s; noemp; eauto.
    + s; noemp; eauto.
    + s; noemp; eauto.
Grab Existential Variables.
  exact 0.
Qed.

Lemma cps_correct_unit: forall F G,
  cps_correct F G Eunit Tunit.
Proof.
  eexists; ii; s; s in UNQ; s in DJ; destruct (ka _ _) eqn:K; s; s in UNQ; s in DJ.
  instantiate (1:=1); prepare; eapply pesim_dtsteps; [by eauto| | |by eauto|].
  - eapply dtsteps_some.
    + apply mid_wdet; s; auto.
    + inst_pairs; do 3 econs.
    + apply dtsteps_none.
  - rr; eauto 10.
  - simpl cplus at 2; noemp; i; eapply stutter; eauto.
    s; noemp; fold (snd (n0, e)); rewrite <-K.
    exploit KSIM; clean; eauto.
    + econs; econs; s; eauto; econs; s; eauto.
    + i; eapply list_assoc_cons_fresh; eauto.
    + rewrite K; eauto.
    + instantiate (1 := next n); clean.
      rewrite K; i.
      apply list_assoc_cons_or in LU; des; subst; eauto.
    + i; rewrite ->mid_csplit, src_csplit; apply pesim_release; rr; eauto.
Qed.

Lemma cps_correct_nat: forall F G n,
  cps_correct F G (Enat n) Tnat.
Proof.
  eexists; ii; s; s in UNQ; s in DJ; destruct (ka _ _) eqn:K; s; s in UNQ; s in DJ.
  instantiate (1:=1); prepare; eapply pesim_dtsteps; [by eauto| | |by eauto|].
  - eapply dtsteps_some.
    + apply mid_wdet; s; auto.
    + inst_pairs; do 3 econs.
    + apply dtsteps_none.
  - rr; eauto 10.
  - simpl cplus at 2; noemp; i; eapply stutter; eauto.
    s; noemp; fold (snd (n1, e)); rewrite <-K.
    exploit KSIM; eauto.
    + instantiate (1 := Vnat n); instantiate (1 := IVnat n).
      econs; econs; s; eauto; econs; s; eauto.
    + i; eapply list_assoc_cons_fresh; eauto.
    + clean; rewrite K; eauto.
    + instantiate (1 := next n0); clean.
      rewrite K; i.
      apply list_assoc_cons_or in LU; des; subst; eauto.
    + i; rewrite ->mid_csplit, src_csplit; apply pesim_release; rr; eauto.
Qed.

Lemma cps_correct_input: forall F G,
  cps_correct F G Einput Tnat.
Proof.
  eexists; ii; s; s in UNQ; s in DJ; destruct (ka _ _) eqn:K; s; s in UNQ; s in DJ.
  pfold; econs; i; esplits; eauto.
  invc CNF; rrdes MA; destruct csa, cfa.
  invc XTRa XTRb; destruct e0; vauto; s in Ec; noemp; clarify.
  eapply esim_step; auto; i.
  inv ST; [|by exfalso; eapply STK, IRinput]; invc STEP.
  
  destruct csb as [[]]; rrdes MB; destruct e, e0; vauto.
  s in Ec; noemp; vauto.
  exists (h0, Just ((inl (inr n), IVnat n1) :: enva, e')); ex @cemp.
  
  evar (evt: event); evar (mbt: @mach src_lsi).
  assert (ST': step src_lsi evt (h ## x0, env, app_cont kb Einput) mbt).
  { unfold evt, mbt; inst_pairs; econs; eauto. }
  instantiate (1:=n1) in (Value of evt).
  unfold evt, mbt in *; clear evt mbt.
  exists (h, Just env, Just (app_cont kb (Enat n1))); ex @cemp; eexists.
  exists (h ## x0, env, app_cont kb (Enat n1)); eexists; exists false.

  splits; eauto.
  - left; econs; eauto; eapply gres_incl_id; left.
    rewrite ->mid_csplit, src_csplit; apply pesim_release; rr; eauto.
    fold (snd (n0, e')); rewrite <-K; eapply KSIM; eauto.
    + repeat econs; vauto.
    + i; eapply list_assoc_cons_fresh; eauto.
    + clean; rewrite K; eauto.
    + clean; rewrite K; i.
      apply list_assoc_cons_or in LU; des; subst; eauto.
  - econs; [by eauto| | |]; s; noemp; rr; eauto 10.
Grab Existential Variables.
  exact 0. exact 0.
Qed.

Lemma cps_correct_new: forall F G e ty
    (IH: cps_correct F G e ty),
  cps_correct F G (Enew e) (Tref ty).
Proof.
  i; rrdes IH; eexists; i; s; s in UNQ; s in DJ.
  change (app_cont kb _) with (app_cont (SKnew::kb) (subst_exp denvb e)).
  instantiate (1 := _ + _); rewrite <-plus_assoc.
  eapply IH; clear IH; eauto; ii; subst; s; s in UNQ0; s in DJ0.
  destruct (ka _ _) eqn:K; s; s in UNQ0; s in DJ0.
  prepare; s; s in RLa; noemp.
  pfold; econs; i; esplits; auto.
  invc CNF; des; rrdes MA MB.
  invc XTRa XTRb; noemp; s in Ec; s in Ec0; clarify.
  apply esim_step; auto; i.
  invc ST; last (by exfalso; hexploit (pmfinite_alloc va PV0); eauto
                  ; i; des; eapply STK; econs; eauto).
  invc STEP.
  hexploit (pmfinite_alloc vb PV1 FIN1); eauto; i; des.
  destruct s' as [[[]]].
  rewrite pmplus_assoc in PV2; hexploit (pmvalid_plusD _ _ PV2); i; des.
  rewrite pmplus_assoc in X; hexploit (pmvalid_plusD _ _ X); i; des.
  rewrite <-pmplus_assoc in *.
  do 6 econs.
  exists (s0, s1, refst_allocP s2 (tsubst_name tsub ty) loc x2 
                               (confrel_newa HR X0) (confrel_newb HR X2), s3).
  esplits; first (by eauto); [left; econs; eauto| |]; last first; last 1 first.
  - left; inst_pairs; eapply Rok, red_ctx; eauto.
  - s; unfold refst_alloc; eauto.
  - econs; eauto; noemp; rr; inst_pairs; s; noemp; esplits.
    + eauto.
    + rewrite pmplus_assoc; eauto.
    + rewrite -pmplus_assoc; eauto.
    + rewrite -pmplus_assoc; eauto.
    + eauto.
    + rewrite pmplus_assoc; eauto.
    + rewrite -pmplus_assoc; eauto.
    + rewrite -pmplus_assoc; eauto.
  - eapply gres_incl_id; left.
    rewrite ->mid_csplit, src_csplit.
    apply pesim_gknow; i.
    apply pesim_release; last first; rr; eauto.
    + rewrite LEx in GET; clarify.
      eapply confrel_alloc; first 1 last; first (eapply confrel_mon); eauto.
      s; eauto.
    + fold (snd (n1, e')); rewrite <-K.
      eapply KSIM; (safe eauto); first 4 last; eauto.
      * i; eapply LE0 in IN; eapply list_assoc_cons_fresh; eauto. 
      * clean; rewrite K; eauto.
      * clean; rewrite K; i.
        apply list_assoc_cons_or in LU; des; subst; eauto.
      * eapply extpub_trans; [|by eauto]; s; unfold refst_alloc; eauto.
      * eapply ext_trans; [|by eauto]; s; unfold refst_alloc; eauto.
      * econs; simpl tsubst_name; econs; rr; esplits; s; rr; eauto.
Grab Existential Variables.
  exact 0.
Qed.

Lemma cps_correct_read: forall F G e ty
    (IH: cps_correct F G e (Tref ty)),
  cps_correct F G (Eread e) ty.
Proof.
  i; rrdes IH; eexists; i; s; s in UNQ; s in DJ.
  change (app_cont kb _) with (app_cont (SKread::kb) (subst_exp denvb e)).
  instantiate (1 := _ + _); rewrite <-plus_assoc.
  eapply IH; clear IH; eauto; ii; subst; s; s in UNQ0; s in DJ0.
  destruct (ka _ _) eqn:K; s; s in UNQ0; s in DJ0.
  invc RV REL; [|by destruct ty0 as [[]]; vauto].
  destruct s' as [[[]]]; rrdes VR; clean; subst.
  prepare; s; s in RLa; noemp.
  dup WHY0; rrdes WHY0; clarify.
  hexploit RELV; eauto; i; des.
  eapply pesim_dstep.
  - rrdes RLa; rr; eauto 10.
  - eapply mid_wdet; s; eauto.
  - inst_pairs; econs; econs; eauto.
  - rr; eauto 10.
  - i; invc XTR; dup RLb; rrdes RLb; s in Ec; clarify; esplits.
    + rewrite Eb; inst_pairs.
      eapply Rok, red_ctx; eauto using pmread_plus.
    + instantiate (1:=(_,Just _,Just _)); rr; s; eauto 10.
    + rewrite ->mid_csplit, src_csplit.
      apply pesim_release; eauto.
      fold (snd (n1, e0)); rewrite <-K.
      exploit KSIM; last 8 first
      ; first (by eapply vclos_lift; eauto)
      ; eauto.
      * i; eapply LE0 in IN.
        i; eapply list_assoc_cons_fresh; eauto.
      * clean; rewrite K; eauto.
      * clean; rewrite K; i.
        apply list_assoc_cons_or in LU; des; subst; eauto.
      * instantiate (1:=0); eauto.
      * rr; eauto.
      * rr; eauto.
Qed.

Lemma cps_correct_output: forall F G e
    (IH: cps_correct F G e Tnat),
  cps_correct F G (Eoutput e) Tunit.
Proof.
  i; rrdes IH; eexists; i; s; s in UNQ; s in DJ.
  change (app_cont kb _) with (app_cont (SKoutput::kb) (subst_exp denvb e)).
  instantiate (1 := _ + _); rewrite <-plus_assoc.
  eapply IH; clear IH; eauto; ii; subst; s; s in UNQ0; s in DJ0.
  destruct (ka _ _) eqn:K; s; s in UNQ0; s in DJ0.
  invc RV REL; [|by destruct ty as [[]]; vauto]; rrdes CB; invc CA SVQ.
  destruct vb, e_of; vauto.
  prepare; s; s in RLa; noemp.
  eapply dolet; try unfold eval; eauto.
  eapply pesim_dstep.
  - rrdes RLa; rr; eauto 10.
  - eapply mid_wdet; s; eauto.
  - instantiate (1:= (_,_)); econs; econs.
    eapply list_assoc_cons_fresh; eauto.
  - rr; eauto 10.
  - i; invc XTR; dup RLb; rrdes RLb; s in Ec; clarify; esplits.
    + inst_pairs; econs; eauto.
    + instantiate (1:=(_,Just _,Just _)); rr; s; eauto 10.
    + rewrite ->mid_csplit, src_csplit; apply pesim_release; rr; eauto.
      fold (snd (n1, e0)); rewrite <-K.
      exploit KSIM; last 8 first
      ; first (by econs; econs; s; eauto; econs; eauto)
      ; eauto.
      * i; eapply LE0 in IN.
        i; eapply list_assoc_cons_fresh; eauto.
      * clean; rewrite K; eauto.
      * clean; rewrite K; i.
        apply list_assoc_cons_or in LU; des; subst; eauto.
Qed.

Lemma cps_correct_ifnz: forall F G e el er ty
    (IH: cps_correct F G e Tnat)
    (IHl: cps_correct F G el ty)
    (IHr: cps_correct F G er ty),
  cps_correct F G (Eifnz e el er) ty.
Proof.
  i; rrdes IH IHl IHr; eexists; i; s; s in UNQ; s in DJ.
  destruct (cps e _ _ _) eqn:E; rewrite <-E.
  change (app_cont kb _) with
    (app_cont (SKifnz (subst_exp denvb el) (subst_exp denvb er) :: kb)
              (subst_exp denvb e)).
  instantiate (1 := _ + _); rewrite <-plus_assoc.
  eapply IH; clear IH; eauto; ii; subst; s; s in UNQ; s in DJ; [by rewrite E| |].
  { rewrite E in IN. eauto. }
  s in UNQ0; s in DJ0.
  destruct (cps el _ _ _) eqn:El; destruct (cps er _ _ _) eqn:Er; s; s in UNQ0; s in DJ0.
  inv RV REL; [|by destruct ty0 as [[]]; vauto]; rrdes CB; invc CA SVQ. 
  instantiate (1 := 1 + max i'0 i'1); destruct n4. 
  - 
    prepare; eapply pesim_dstep; [by eauto| | | |].
    + eapply mid_wdet; s; eauto.
    + inst_pairs; rr; eauto.
    + rr; eauto 10.
    + { i; invc XTR; dup RLb; rrdes RLb; s in Ec; clarify; esplits.
      - inst_pairs; econs; eauto.
      - instantiate (1:=(_,Just _,Just _)); rr; s; eauto 10.
      - noemp; rewrite ->mid_csplit, src_csplit; apply pesim_release; rr; eauto.
        apply pesim_gknow; i; fold (snd (n3, e2)); rewrite <-Er.
        
        eapply pesim_wfo_le; [eapply IHr|]; eauto.
        + i; hexploit RSV; eauto; i; des.
          erewrite cpsw_ext; eauto 20 using vclos_mon, (gk_mon GK0).
        + i; hexploit RDV; eauto; i; des.
          eauto 10 using vclos_mon, (gk_mon GK0).
        + rewrite Er; s; eapply uniq_sublist; eauto.
        + rewrite Er; eauto using in_or_app.
        + ii; eapply paco10_mon; eauto.
          eapply KSIM; eauto using ext_trans.
        + apply wfo_nat_le_mon, lt_le_weak. 
          rhs (rewrite plus_comm); rewrite plus_assoc; s.
          rewrite plus_comm; rewrite <-plus_assoc.
          eapply plus_lt_compat_l. eapply le_n_S, Max.le_max_r.
      }
  - 
    prepare; eapply pesim_dstep; [by eauto| | | |].
    + eapply mid_wdet; s; eauto.
    + inst_pairs; rr; eauto.
    + rr; eauto 10.
    + { i; invc XTR; dup RLb; rrdes RLb; s in Ec; clarify; esplits.
      - inst_pairs; econs; eauto.
      - instantiate (1:=(_,Just _,Just _)); rr; s; eauto 10.
      - noemp; rewrite ->mid_csplit, src_csplit; apply pesim_release; rr; eauto.
        apply pesim_gknow; i; fold (snd (n2, e1)); rewrite <-El.
        eapply pesim_wfo_le; [eapply IHl|]; eauto.
        + i; hexploit RSV; eauto; i; des.
          erewrite cpsw_ext; eauto 20 using vclos_mon, (gk_mon GK0).
        + i; hexploit RDV; eauto; i; des.
          eauto 10 using vclos_mon, (gk_mon GK0).
        + rewrite El; s; eapply uniq_sublist; eauto.
        + rewrite El; eauto using in_or_app.
        + ii; eapply paco10_mon; eauto.
          eapply KSIM; eauto using ext_trans.
        + apply wfo_nat_le_mon, lt_le_weak.
          rhs (rewrite plus_comm); rewrite plus_assoc; s.
          rewrite plus_comm; rewrite <-plus_assoc.
          apply plus_lt_compat_l; eapply le_n_S, Max.le_max_l.
      }
Qed.

Lemma ms_refconfrel_write: forall g s ty la lb va vb ha hb
    (RCR: ms_refconfrel g s ({la |-> va} ## ha, Nothing)
                            ({lb |-> vb} ## hb, Nothing, Nothing))
    (RL: snd s ty la lb)
    va' vb' (VCL: vclos ms_gw s (g s) ty va' vb'),
  ms_refconfrel g s ({la |-> va'} ## ha, Nothing)
                    ({lb |-> vb'} ## hb, Nothing, Nothing).
Proof.
  i; rrdes RCR; clarify.
  rr; esplits; eauto using pmvalid_plus_cell.
  - i; eapply pmread_plusD in ALC; des.
    + eapply pmread_cellD in ALC; des; clarify.
      eauto using pmread_plus, pmread_cell.
    + eapply DOMa; eauto using pmread_plus, pmplus_comm.
  - i; eapply pmread_plusD in ALC; des.
    + eapply pmread_cellD in ALC; des; clarify.
      eauto using pmread_plus, pmread_cell.
    + eapply DOMb; eauto using pmread_plus, pmplus_comm.
  - i; dup REG; apply RELV in REG; des.
    hexploit (refst_bij REG0 RL); i; des.
    apply pmread_plusD in RDa; apply pmread_plusD in RDb; des.
    + apply pmread_cellD in RDa; apply pmread_cellD in RDb; des; subst.
      specialize (X eq_refl); specialize (X0 eq_refl); des; subst; clean.
      destruct s; esplits; eauto using pmread_plus, pmread_cell, pmvalid_plus_cell.
    + exfalso.
      apply pmread_cellD in RDb; des; subst.
      specialize (X0 eq_refl); des; subst; clean.
      rrdes RDa; subst.
      apply RELV in RL; des.
      rewrite pmplus_assoc in VLDa.
      apply pmvalid_plus_l in VLDa.
      apply pmvalid_cell_cell in VLDa; vauto.
    + exfalso.
      apply pmread_cellD in RDa; des; subst.
      specialize (X eq_refl); des; subst; clean.
      rrdes RDb; subst.
      apply RELV in RL; des.
      rewrite pmplus_assoc in VLDb.
      apply pmvalid_plus_l in VLDb.
      apply pmvalid_cell_cell in VLDb; vauto.
    + eauto 10 using pmread_plus, pmplus_comm, pmvalid_plus_cell.
Qed.

Lemma ms_confrel_shape: forall g (s: state ms_gwsts) ha hb
    (WH: confrel ms_gw g s ha hb),
  exists ha' hb',
    ha = (ha', Nothing) /\
    hb = (hb', Just (snd (fst s)), Nothing).
Proof.
  i; rrdes WH WHX WHY; rrdes CPa CPb; subst; s; noemp; eauto.
Qed.

Lemma ms_confrel_write: forall g s ty la lb va vb ha hb
    (CR: confrel cpsw g s ({la |-> va} ## ha, Nothing)
                          ({lb |-> vb} ## hb, Just (proj_env s), Nothing))
    (RL: snd (fst s) ty la lb)
    va' vb' (VCL: vclos cpsw s (g s) ty va' vb'),
  confrel cpsw g s ({la |-> va'} ## ha, Nothing)
                   ({lb |-> vb'} ## hb, Just (proj_env s), Nothing).

Proof.
  i; rrdes CR WHY.
  hexploit ms_confrel_shape; eauto; i; des; clarify; noemp; clarify.
  rrdes WHX WHX0 CPa CPb; clarify; noemp; clarify.
  destruct hby as [[]]; s in HPB; destruct e, e0; vauto; noemp; subst.
  rr; esplits; rr; eauto; noemp; eauto; destruct s.
  esplits; first (by rr; esplits; rr; eauto); noemp; last first
  ; eauto using ms_refconfrel_write, vclos_unlift.
  - instantiate (1:=(_,Nothing,Nothing)); s; noemp; eauto.
  - eapply ms_refconfrel_write; eauto using vclos_unlift.
Qed.

Lemma cps_correct_refeq: forall F G ty el er
    (IHl: cps_correct F G el (Tref ty))
    (IHr: cps_correct F G er (Tref ty)),
  cps_correct F G (Erefeq el er) Tnat.
Proof.
  i; rrdes IHl IHr; eexists; i; s; s in UNQ.
  change (app_cont kb _) with (app_cont (SKrefeqL (subst_exp denvb er) :: kb) (subst_exp denvb el)).
  instantiate (1 := _ + _); rewrite <-plus_assoc.
  eapply IHl; eauto; clear IHl.
  ii; subst; s; s in UNQ0; s in DJ0.
  change (app_cont kb _) with (app_cont (SKrefeqR vb :: kb) (subst_exp denvb er)).
  instantiate (1 := _ + _); rewrite <-plus_assoc.
  eapply pesim_gknow; i.
  eapply IHr; (safe eauto); clear IHr; last 1 first.
  { eauto. }
  { i; hexploit RSV; eauto; i; des.
    erewrite cpsw_ext; eauto 20 using vclos_mon, (gk_mon GK0). }
  { i; hexploit RDV; eauto; i; des.
    eauto 10 using vclos_mon, (gk_mon GK0). }
  ii; subst; s; s in UNQ1; s in DJ1.
  destruct (ka _ _) eqn:K; s; s in UNQ1; s in DJ1.
  invc RV REL; [|by destruct ty0 as [[]]; vauto].
  inv RV0 REL; [|by destruct ty0 as [[]]; vauto].
  destruct VR as [mla [sla [Ha1 [Ha2 Ha3]]]].
  destruct VR0 as [mlb [slb [Hb1 [Hb2 Hb3]]]].
  prepare; eapply pesim_dstep; [by eauto| | | |].
  + eapply mid_wdet; s; eauto.
  + inst_pairs; econs; instantiate (2:= _::_).
    eapply IRlet; eapply LE2 in LEx; s; rewrite LEx LEx0; by eauto.
  + rr; by eauto 10. 
  + i; rrdes RLb; invc XTR; s in Ec; clarify; noemp; esplits.
    * instantiate (1 := (_,_,_)); eapply Rok, red_ctx; eauto. 
      move: Ha2 Hb2 => -> {vb} ->.
      rrdes WHY0; clean; clarify.
      hexploit RELV; eauto; i; des.
    * rr; esplits; [|eauto| |]; eauto.
      instantiate (1:=(_,Just _,Just _)); s; by eauto.
    * rewrite ->mid_csplit, src_csplit.
      { apply (@pesim_release _ _ cpsw); eauto.
        fold (snd (n2, e)); rewrite <-K; eauto.
        instantiate (1:=0); eapply KSIM; (safe (eauto using ext_trans)); first 2 last.
        - by eauto.
        - by eauto.
        - i; eapply list_assoc_cons_fresh; by eauto.
        - clean; rewrite K; by eauto.
        - clean; rewrite K; i.
          apply list_assoc_cons_or in LU; des; subst; by eauto.
        - econs; econs; last first.
          + econs; s; eauto; s; by eauto.
          + s; by eauto.
        - rewrite /lookup. rewrite eqdec_true. 
          assert (Ha3': snd (fst s'0) (tsubst_name tsub ty) mla sla).
          { destruct EXT1. destruct H. eauto using H0. }
          move: (refst_bij Hb3 Ha3') => [F1 F2].
          case: (eq_nat_dec mla mlb); case: (eq_nat_dec sla slb); intros H1 H2; try by eauto.
          * by move: (F1 H2) => [_].
          * by move: (F2 H1) => [_].
        - rr; by eauto.
        - rr; by eauto.
      }
Qed.

Lemma cps_correct_write: forall F G el er ty
    (IHl: cps_correct F G el (Tref ty))
    (IHr: cps_correct F G er ty),
  cps_correct F G (Ewrite el er) Tunit.
Proof.
  i; rrdes IHl IHr; eexists; i; s; s in UNQ.
  destruct (cps el _ _ _) eqn:El; rewrite <-El.
  change (app_cont kb _) with
    (app_cont (SKwriteL (subst_exp denvb er) :: kb)
              (subst_exp denvb el)).
  instantiate (1 := _ + _); rewrite <-plus_assoc.
  eapply IHl; clear IHl; eauto; ii; subst; s; s in UNQ
  ; [by rewrite El; eauto|].
  s in UNQ0; destruct (cps er _ _ _) eqn:Er; rewrite <-Er; s; s in UNQ0.
  inv RV REL; [|by destruct ty0 as [[]]; vauto].
  destruct s' as [[[]]]; rrdes VR; clean; subst.
  change (app_cont kb _) with
    (app_cont (SKwriteR vb :: kb) (subst_exp denvb er)).
  eapply pesim_gknow; i.
  instantiate (1 := _ + _); rewrite <-plus_assoc.
  eapply IHr; (safe eauto); clear IHr; last by eauto.
  { i; hexploit RSV; eauto; i; des.
    erewrite cpsw_ext; eauto 20 using vclos_mon, (gk_mon GK0). }
  { i; hexploit RDV; eauto; i; des.
    eauto 10 using vclos_mon, (gk_mon GK0). }
  { rewrite Er; eauto. }
  ii; subst; s; s in UNQ0; s in DJ0; inst; eauto.
  s in DJ1; s in UNQ1; destruct (ka _ _) eqn:K; s; s in UNQ1; s in DJ1.
  prepare; s; s in RLa; noemp.
  eapply dolet; try unfold eval; eauto.
  invc REL VR; des; clarify; rewrite Eb0.
  rrdes WHY0; clean; clarify.
  eapply EXT1 in REG0.
  hexploit RELV; eauto; i; des.
  rrdes RDa RDb; subst.
  eapply pesim_dstep.
  - rrdes RLa; rr; eauto 10.
  - eapply mid_wdet; s; eauto.
  - inst_pairs; econs; econs.
    + eapply LE2 in LEx.
      eapply list_assoc_cons_fresh; eauto.
    + eapply list_assoc_cons_fresh; eauto.
  - rr; esplits; eauto using pmvalid_plus_cell, pmfinite_plus_cell.
  - i; invc XTR; dup RLb; rrdes RLb; s in Ec; clarify; esplits.
    + rewrite -pmplus_assoc; inst_pairs; econs; eauto.
    + instantiate (1:=(_,Just _,Just _)); rr; s; esplits; safe eauto.
      * eauto.
      * rewrite pmplus_assoc; eauto.
      * rewrite <-!pmplus_assoc in *; eauto using pmvalid_plus_cell.
      * rewrite <-!pmplus_assoc in *; eauto using pmfinite_plus_cell.
    + rewrite ->mid_csplit, src_csplit.
      apply pesim_release; eauto using ms_confrel_write.
      fold (snd (n4, e1)); rewrite <-K.
      exploit KSIM; last 8 first
      ; first (by econs; econs; s; eauto; econs; eauto)
      ; eauto.
      * i; eapply list_assoc_cons_fresh; eauto.
      * clean; rewrite K; eauto.
      * clean; rewrite K; i.
        apply list_assoc_cons_or in LU; des; subst; eauto.
      * eauto.
      * eauto using ext_trans.
      * rr; eauto.
      * rr; eauto.
Qed.  

Lemma cps_correct_case: forall F G e el er ty tyl tyr x
    (IH: cps_correct F G e (Tsum tyl tyr))
    (IHl: cps_correct F ((x, tyl)::G) el ty)
    (IHr: cps_correct F ((x, tyr)::G) er ty),
  cps_correct F G (Ecase e x el er) ty.
Proof.
  i; rrdes IH IHl IHr; eexists; i; s; s in UNQ.
  destruct (cps e _ _ _) eqn:E; rewrite <-E.
  replace (app_cont kb _) with
    (app_cont (SKcase x (subst_exp [Evar x//x]>>denvb el) (subst_exp [Evar x//x]>>denvb er) :: kb)
              (subst_exp denvb e)); eauto. 
  instantiate (1 := _ + _); rewrite <-plus_assoc.
  eapply IH; clear IH; eauto; ii; subst; s; s in UNQ; s in DJ; [by rewrite E|].
  s in UNQ0; s in DJ0.
  destruct (cps el _ _ _) eqn:El; destruct (cps er _ _ _) eqn:Er; s; s in UNQ0; s in DJ0.
  instantiate (1 := 1 + max i'0 i'1).
  inv RV REL; [| |by destruct ty0 as [[]]; vauto]; rrdes CB; invc CA SVQ. 
  - 
    prepare; eapply pesim_dstep; [by eauto| | | |].
    + eapply mid_wdet; s; eauto.
    + inst_pairs; econs; eapply IRcasel; eauto.
    + rr; eauto 10.
    + { i; invc XTR; dup RLb; rrdes RLb; s in Ec; clarify; esplits.
      - inst_pairs; econs; eauto.
      - instantiate (1:=(_,Just _,Just _)); rr; s; eauto 10.
      - noemp; rewrite ->mid_csplit, src_csplit.
        apply pesim_release; rr; eauto; apply pesim_gknow; i.
        fold (snd (n2, e1)); rewrite <-El.
        rewrite <-subexp_tmext_l, tmext_lift; eauto using subclo_weakly.
        eapply pesim_wfo_le; [eapply IHl|]; eauto.
        + i; hexploit RSV; eauto; i; des.
          rewrite list_assoc_cons_neq; vauto.
          erewrite cpsw_ext; eauto 20 using vclos_mon, (gk_mon GK0).
        + i; apply list_assoc_cons_or in IN; des; subst.
          * rewrite !tmext_eq list_assoc_cons_eq; esplits; eauto.
          * rewrite !tmext_neq; eauto.
            hexploit RDV; eauto; i; des.
            esplits; eauto 10 using vclos_mon, (gk_mon GK0).
            eapply LE0 in INa.
            eapply list_assoc_cons_fresh; eauto.
        + rewrite El; eapply uniq_appD in UNQ1; des; eauto.
        + clean; rewrite El; i.
          apply list_assoc_cons_or in LU; des; subst; eauto using in_or_app.
        + rewrite <-tmext_lift; eapply weakclo_tmext; eauto.
          ii; exfalso; eauto using ms_gw_vclos_closed.
        + i; apply in_consD in IN; des; subst.
          * rewrite tmext_eq in FH; eauto using ms_gw_vclos_closed.
          * rewrite tmext_neq in FH; eauto.
        + ii; eapply paco10_mon; eauto.
          eapply KSIM; eauto using ext_trans.
          i; eapply LE2.
          eapply LE0 in IN.
          eapply list_assoc_cons_fresh; eauto.
        + apply wfo_nat_le_mon, lt_le_weak.
          rhs (rewrite plus_comm); rewrite plus_assoc; s.
          rewrite plus_comm; rewrite <-plus_assoc.
          apply plus_lt_compat_l; eapply le_n_S, Max.le_max_l.
      }
  - 
    prepare; eapply pesim_dstep; [by eauto| | | |].
    + eapply mid_wdet; s; eauto.
    + inst_pairs; econs; eapply IRcaser; eauto.
    + rr; eauto 10.
    + { i; invc XTR; dup RLb; rrdes RLb; s in Ec; clarify; esplits.
      - inst_pairs; econs; eauto.
      - instantiate (1:=(_,Just _,Just _)); rr; s; eauto 10.
      - noemp; rewrite ->mid_csplit, src_csplit.
        apply pesim_release; rr; eauto; apply pesim_gknow; i.
        fold (snd (n3, e2)); rewrite <-Er.
        rewrite <-subexp_tmext_l, tmext_lift; eauto using subclo_weakly.
        eapply pesim_wfo_le; [eapply IHr|]; eauto.
        + i; hexploit RSV; eauto; i; des.
          rewrite list_assoc_cons_neq; vauto.
          erewrite cpsw_ext; eauto 20 using vclos_mon, (gk_mon GK0).
        + i; apply list_assoc_cons_or in IN; des; subst.
          * rewrite !tmext_eq list_assoc_cons_eq; esplits; eauto.
          * rewrite !tmext_neq; eauto.
            hexploit RDV; eauto; i; des.
            esplits; eauto 10 using vclos_mon, (gk_mon GK0).
            eapply LE0 in INa.
            eapply list_assoc_cons_fresh; eauto.
        + rewrite Er; eapply uniq_appD in UNQ1; des; eauto.
        + rewrite Er; i.
          apply list_assoc_cons_or in LU; des; subst; eauto using in_or_app.
        + rewrite <-tmext_lift; eapply weakclo_tmext; eauto.
          i; exfalso; eauto using ms_gw_vclos_closed.
        + i; apply in_consD in IN; des; subst.
          * rewrite tmext_eq in FH; eauto using ms_gw_vclos_closed.
          * rewrite tmext_neq in FH; eauto. 
        + ii; eapply paco10_mon; eauto.
          eapply KSIM; eauto using ext_trans.
          i; eapply LE2.
          eapply LE0 in IN.
          eapply list_assoc_cons_fresh; eauto.
        + apply wfo_nat_le_mon, lt_le_weak.
          rhs (rewrite plus_comm); rewrite plus_assoc; s.
          rewrite plus_comm; rewrite <-plus_assoc.
          apply plus_lt_compat_l; eapply le_n_S, Max.le_max_r.
      }
Qed.



Lemma cps_correct_fst: forall F G e tya tyb
    (IH: cps_correct F G e (Tprod tya tyb)),
  cps_correct F G (Efst e) tya.
Proof.
  i; rrdes IH; eexists; i; s; s in UNQ.
  change (app_cont kb _) with (app_cont (SKfst::kb) (subst_exp denvb e)).
  eapply IH; eauto; clear IH; ii; subst; s; s in UNQ0; s in DJ0.
  destruct (ka _ _) eqn:K; s; s in UNQ0; s in DJ0.
  invc RV REL; [|by destruct ty as [[]]; vauto]; rrdes CB; invc CA SVQ.
  prepare; eapply pesim_dstep; [by eauto| | | |].
  + eapply mid_wdet; s; eauto.
  + inst_pairs; econs; instantiate (2:= _::_).
    eapply IRlet; unfold eval; setoid_rewrite LEx; s; eauto.
  + rr; eauto 10.
  + i; rrdes RLb; invc XTR; s in Ec; clarify; noemp; esplits.
    * inst_pairs; eapply Rok, red_ctx; eauto.
    * rr; esplits; [|eauto| |]; eauto.
      instantiate (1:=(_,Just _,Just _)); s; eauto.
    * rewrite ->mid_csplit, src_csplit.
      { apply (@pesim_release _ _ cpsw); eauto.
        eapply KSIM; eauto.
        - s; fold (snd (n1, e0)); rewrite <-K; eauto.
        - i; eapply list_assoc_cons_fresh; eauto.
        - clean; rewrite K; eauto.
        - clean; rewrite K; i.
          apply list_assoc_cons_or in LU; des; subst; eauto.
        - rr; eauto.
        - rr; eauto.
      }
Qed.

Lemma cps_correct_snd: forall F G e tya tyb
    (IH: cps_correct F G e (Tprod tya tyb)),
  cps_correct F G (Esnd e) tyb.
Proof.
  i; rrdes IH; eexists; i; s; s in UNQ.
  change (app_cont kb _) with (app_cont (SKsnd::kb) (subst_exp denvb e)).
  eapply IH; eauto; clear IH; ii; subst; s; s in UNQ0; s in DJ0.
  destruct (ka _ _) eqn:K; s; s in UNQ0; s in DJ0.
  invc RV REL; [|by destruct ty as [[]]; vauto]; rrdes CB; invc CA SVQ.
  prepare; eapply pesim_dstep; [by eauto| | | |].
  + eapply mid_wdet; s; eauto.
  + inst_pairs; econs; instantiate (2:= _::_).
    eapply IRlet; unfold eval; setoid_rewrite LEx; s; eauto.
  + rr; eauto 10.
  + i; rrdes RLb; invc XTR; s in Ec; clarify; noemp; esplits.
    * inst_pairs; eapply Rok, red_ctx; eauto.
    * rr; esplits; [|eauto| |]; eauto.
      instantiate (1:=(_,Just _,Just _)); s; eauto.
    * rewrite ->mid_csplit, src_csplit.
      { apply (@pesim_release _ _ cpsw); eauto.
        eapply KSIM; eauto.
        - s; fold (snd (n1, e0)); rewrite <-K; eauto.
        - i; eapply list_assoc_cons_fresh; eauto.
        - clean; rewrite K; eauto.
        - clean; rewrite K; i.
          apply list_assoc_cons_or in LU; des; subst; eauto.
        - rr; eauto.
        - rr; eauto.
      }
Qed.

Lemma cps_correct_inl: forall F G e tya tyb
    (IH: cps_correct F G e tya),
  cps_correct F G (Einl e) (Tsum tya tyb).
Proof.
  i; rrdes IH; eexists; i; s; s in UNQ.
  change (app_cont kb _) with (app_cont (SKinl::kb) (subst_exp denvb e)).
  instantiate (1 := _ + _); rewrite <-plus_assoc.
  eapply IH; clear IH; eauto; ii; subst; s; s in UNQ0; s in DJ0.
  destruct (ka _ _) eqn:K; s; s in UNQ0; s in DJ0.
  invc RV.
  prepare; eapply pesim_dtsteps; [by eauto| | |by eauto|].
  + eapply dtsteps_some; [by eapply mid_wdet; s; eauto| |].
    * inst_pairs; econs; instantiate (2:= _::_).
      eapply IRlet; unfold eval; setoid_rewrite LEx; s; eauto.
    * eapply dtsteps_none.
  + rr; eauto 10.
  + simpl cplus at 2; noemp.
    i; eapply stutter; eauto.
    rewrite ->mid_csplit, src_csplit; apply (@pesim_release _ _ cpsw); rr; eauto.
    fold (snd (n1, e0)); rewrite <-K; instantiate (1:=0).
    exploit KSIM; last (by eauto); [| | | |M| | | | |]; Mskip eauto.
    * econs; eapply vclgen_inl; s; eauto; econs; [by econs|].
      ii; s in FH; eapply ms_gw_vclos_closed; eauto.
    * eauto.
    * i; eapply list_assoc_cons_fresh; eauto.
    * clean; rewrite K; eauto.
    * clean; rewrite K; i.
      apply list_assoc_cons_or in LU; des; subst; eauto.
Qed.

Lemma cps_correct_inr: forall F G e tya tyb
    (IH: cps_correct F G e tyb),
  cps_correct F G (Einr e) (Tsum tya tyb).
Proof.
  i; rrdes IH; eexists; i; s; s in UNQ.
  change (app_cont kb _) with (app_cont (SKinr::kb) (subst_exp denvb e)).
  instantiate (1 := _ + _); rewrite <-plus_assoc.
  eapply IH; clear IH; eauto; ii; subst; s; s in UNQ0; s in DJ0.
  destruct (ka _ _) eqn:K; s; s in UNQ0; s in DJ0.
  invc RV.
  prepare; eapply pesim_dtsteps; [by eauto| | |by eauto|].
  + eapply dtsteps_some; [by eapply mid_wdet; s; eauto| |].
    * inst_pairs; econs; instantiate (2:= _::_).
      eapply IRlet; unfold eval; setoid_rewrite LEx; s; eauto.
    * eapply dtsteps_none.
  + rr; eauto 10.
  + simpl cplus at 2; noemp.
    i; eapply stutter; eauto.
    rewrite ->mid_csplit, src_csplit; apply (@pesim_release _ _ cpsw); rr; eauto.
    fold (snd (n1, e0)); rewrite <-K; instantiate (1:=0).
    exploit KSIM; last (by eauto); [| | | |M| | | | |]; Mskip eauto.
    * econs; eapply vclgen_inr; s; eauto; econs; [by econs|].
      ii; s in FH; eapply ms_gw_vclos_closed; eauto.
    * eauto.
    * i; eapply list_assoc_cons_fresh; eauto.
    * clean; rewrite K; eauto.
    * clean; rewrite K; i.
      apply list_assoc_cons_or in LU; des; subst; eauto.
Qed.

Lemma cps_correct_bop: forall F G op el er
    (IHl: cps_correct F G el Tnat)
    (IHr: cps_correct F G er Tnat),
  cps_correct F G (Ebop op el er) Tnat.
Proof.
  i; rrdes IHl IHr; eexists; i; s; s in UNQ.
  change (app_cont kb _) with
    (app_cont (SKbopL op (subst_exp denvb er) :: kb) (subst_exp denvb el)).
  instantiate (1 := _ + _); rewrite <-plus_assoc.
  eapply IHl; eauto; clear IHl.
  ii; subst; s; s in UNQ0; s in DJ0.
  change (app_cont kb _) with (app_cont (SKbopR op vb :: kb) (subst_exp denvb er)).
  instantiate (1 := _ + _); rewrite <-plus_assoc.
  eapply pesim_gknow; i.
  eapply IHr; (safe eauto); clear IHr; last 1 first.
  { eauto. }
  { i; hexploit RSV; eauto; i; des.
    erewrite cpsw_ext; eauto 20 using vclos_mon, (gk_mon GK0). }
  { i; hexploit RDV; eauto; i; des.
    eauto 10 using vclos_mon, (gk_mon GK0). }
  ii; subst; s; s in UNQ1; s in DJ1.
  destruct (ka _ _) eqn:K; s; s in UNQ1; s in DJ1.
  invc RV REL; [|by destruct ty as [[]]; vauto]; rrdes CB; invc CA SVQ.
  inv RV0 REL; [|by destruct ty; destruct flext_of; vauto].
  inv CA; rrdes CB; invc SVQ.
  prepare; eapply pesim_dstep; [by eauto| | | |].
  - eapply mid_wdet; s; eauto.
  - inst_pairs; econs; instantiate (2:= _::_).
    eapply IRlet; eapply LE2 in LEx; s; rewrite LEx LEx0; eauto.
  - rr; eauto 10. 
  - i; rrdes RLb; invc XTR; s in Ec; clarify; noemp; esplits.
    + inst_pairs; eapply Rok, red_ctx; eauto.
    + rr; esplits; [|eauto| |]; eauto.
      instantiate (1:=(_,Just _,Just _)); s; eauto.
    + rewrite ->mid_csplit, src_csplit; apply pesim_release; rr; eauto.
      fold (snd (n2, e)); rewrite <-K; eauto.
      instantiate (1:=0); eapply KSIM; (safe (eauto using ext_trans)); first 2 last.
      * eauto.
      * eauto.
      * i; eapply list_assoc_cons_fresh; eauto.
      * clean; rewrite K; eauto.
      * clean; rewrite K; i.
        apply list_assoc_cons_or in LU; des; subst; eauto.
      * econs; econs; last first; try econs; s; eauto.
      * eauto.
Qed.

Lemma cps_correct_pair: forall F G el er tyl tyr
    (IHl: cps_correct F G el tyl)
    (IHr: cps_correct F G er tyr),
  cps_correct F G (Epair el er) (Tprod tyl tyr).
Proof.
  i; rrdes IHl IHr; eexists; i; s; s in UNQ.
  change (app_cont kb _) with
    (app_cont (SKpairL (subst_exp denvb er) :: kb) (subst_exp denvb el)).
  instantiate (1 := _ + _); rewrite <-plus_assoc.
  eapply IHl; eauto; clear IHl.
  ii; subst; s; s in UNQ0; s in DJ0.
  change (app_cont kb _) with (app_cont (SKpairR vb :: kb) (subst_exp denvb er)).
  instantiate (1 := _ + _); rewrite <-plus_assoc.
  eapply pesim_gknow; i.
  eapply IHr; (safe eauto); clear IHr; last 1 first.
  { eauto. }
  { i; hexploit RSV; eauto; i; des.
    erewrite cpsw_ext; eauto 20 using vclos_mon, (gk_mon GK0). }
  { i; hexploit RDV; eauto; i; des.
    eauto 10 using vclos_mon, (gk_mon GK0). }
  ii; subst; s; s in UNQ1; s in DJ1.
  destruct (ka _ _) eqn:K; s; s in UNQ1; s in DJ1.
  prepare; eapply pesim_dtsteps; [by eauto| | |by eauto|].
  - eapply dtsteps_some.
    + eapply mid_wdet; s; eauto.
    + inst_pairs; econs; instantiate (2:= _::_).
      eapply IRlet; eapply LE2 in LEx; s; rewrite LEx LEx0; s; eauto.
    + eapply dtsteps_none.
  - rr; eauto 10.
  - i; rrdes RLb; invc XTR; s in Ec; clarify; noemp; esplits; first 1 last.
    + right; instantiate (1:=1); instantiate (1:=wfo_nat _); s; splits; eauto.
      econs; eauto; econs; econs; s; eauto.
    + rr; esplits; [|eauto| |]; eauto.
      instantiate (1:=(_,Just _,Just _)); s; eauto.
    + rewrite ->mid_csplit, src_csplit; apply pesim_release; rr; eauto.
      fold (snd (n2, e)); rewrite <-K; eauto.
      eapply KSIM; (safe (eauto using ext_trans)); first 2 last.
      * eauto.
      * eauto.
      * i; eapply list_assoc_cons_fresh; eauto.
      * clean; rewrite K; eauto.
      * clean; rewrite K; i.
        apply list_assoc_cons_or in LU; des; subst; eauto.
      * econs; econs; last first; eauto 10 using vclos_mon, (gk_mon GK0).
        { rr; splits; s; eauto.
          i; des; eauto using ms_gw_vclos_closed.
        }
        { rr; splits; s; eauto. }
      * eauto.
Qed.

Lemma cps_correct_gen: forall F G e ty aa
    (IH: cps_correct F G e (tsubst_bvar 0 (Tname aa) ty))
    (FTF: ftn_typ_l (ty :: map snd F) aa -> False)
    (FTG: ftn_typ_l (ty :: map snd G) aa -> False)
    (HT: hastypepre F G e (tsubst_bvar 0 (Tname aa) ty)),
  cps_correct F G (Egen e) (Tforall ty).
Proof.
  exists 1; i; rrdes IH; s; s in UNQ; s in DJ.
  destruct (cps e _ _ _) eqn:E; destruct (ka _ _) eqn:K; s; s in UNQ; s in DJ.
  prepare; eapply pesim_dtsteps; [by eauto| | |by eauto|].
  - econs; [by apply mid_wdet; s; auto| |by econs].
    inst_pairs; eapply IRok; econs; eauto; econs.
  - rr; eauto 10.
  - i; rrdes XTR RLb; clarify; s in Ec; clarify; esplits.
    2: by right; split; eauto; instantiate (1:=wfo_nat i); econs; eauto
    ; do 2 econs; eauto; s; omega.
    2: by rr; esplits; eauto; inst; instantiate (1:=(_,Just _,Just _)); s; eauto.
    simpl cplus in HR; noemp.
    rewrite ->mid_csplit, src_csplit; eapply (@pesim_release _ _ cpsw); rr; eauto.

    assert (CL': forall a, ~ fv_exp (Vgen (subst_exp denvb e)) a).
    { s; i; eapply subst_exp_closed; last first; eauto using hastype_fv. }

    assert (SIM: vclos cpsw s (g s) (tsubst_name tsub (Tforall ty))
                       (IVclo enva (IEgen (next n) e0))
                       (Vgen (subst_exp denvb e))).
    {
      clear KSIM; simpl tsubst_name. 
      do 3 econs; [by econs|econs; [by econs|by eauto]|].
      apply GK; econs.
      + econs; [by econs; i; inv FH|]; rr; eauto.
      + depgen denvb; depgen enva.
        instantiate (1 := wfo_nat (i' + i)).
        pcofix CIH; i.
        apply pesim_gknow; i.
        invc STK; [by invc APPa RF VR|].
        invc INSTa INSTb; invc RF RK VR; s in VAL; clarify.
        fold (snd (n0, e2)); rewrite <-E.
        apply pesim_gknow_; i.           
        { hexploit (IH ([tyi//aa]>>tsub)).
          - eauto.
          - instantiate (1:=s'); eauto.
          - instantiate (1:= (inr (next n), ka0) :: env).
            i; hexploit RSV; eauto; i; des.
            esplits; last first.
            + eapply vclos_mon; eauto.
              erewrite tsubname_agree.
              * eapply vclos_mon; eauto; i; eapply (gk_mon GK0); eauto.
              * i; clear -FTF IN IN0.
                assert (x <> aa).
                { ii; subst; induction F; vauto.
                  destruct a; apply list_assoc_cons_or in IN; des; subst
                  ; [apply FTF|eapply IHF]; ss; i; des; eauto using FTF.
                }
                rewrite tmext_neq; eauto. 
            + erewrite cpsw_ext; eauto.
            + eauto.
          - i.
            instantiate (1:= denvb).
            instantiate (1:= ren).
            rewrite list_assoc_cons_neq; vauto.
            hexploit RDV; eauto; i; des.
            esplits; eauto.
            eapply vclos_mon; eauto.
            erewrite tsubname_agree.
            * eapply vclos_mon; eauto; i; eapply (gk_mon GK0); eauto.
            * i; clear -FTG IN IN0.
              assert (x0 <> aa).
              { ii; subst; induction G; vauto.
                destruct a; apply list_assoc_cons_or in IN; des; subst
                ; [apply FTG|eapply IHG]; ss; i; des; eauto using FTG.
              }
              rewrite tmext_neq; eauto.
          - simpl map.
            instantiate (1 := fun mx y => (mx, IEret (next n) y)).
            instantiate (1 := next (next n)).
            rewrite E.
            eapply uniq_appD in UNQ1; des; eauto.
          - clean; rewrite E; i.
            apply list_assoc_cons_or in LU; des; subst; eauto 10 using in_or_app.
          - eauto.
          - eauto.
          - eapply cps_ksim_done.
            rewrite list_assoc_cons_eq; eauto.
          - ii; destruct (LEM (x = aa)); subst.
            + rewrite tmext_eq; eauto.
            + rewrite tmext_neq; eauto.
          - i; rewrite tsub_name_bvar in H.
            + s in H; rewrite tmext_eq in H.
              erewrite tsubname_agree in H.
              * eapply paco10_mon; vauto.
              * i; rewrite tmext_neq; eauto.
                ii; subst; apply FTF; s; eauto.
            + ii; destruct (LEM (x = aa)); subst.
              * rewrite tmext_eq; eauto.
              * rewrite tmext_neq; eauto. 
        }
    }
    + unfold cps_ksim in KSIM.
      assert (LE: next (next n) <= n0).
      { exploit cps_uniqvar; eauto; inst; [by s; i; clarify|].
        i; des; rrdes x1; eauto. }
      eapply paco10_mon; [eapply KSIM|by eauto].
      * eauto.
      * vauto.
      * eauto.
      * eauto.
      * eauto.
      * s; instantiate (1:=inr n); instantiate (1:=n0); rewrite K; eauto.
      * eauto.
      * i; eapply list_assoc_cons_fresh; eauto.
      * clean; rewrite K.
        eapply uniq_appD in UNQ1; des; eauto.
      * clean; rewrite K; i.
        apply list_assoc_cons_or in LU; des; subst; eauto 10 using in_or_app.
Qed.

Lemma cps_correct_fix: forall F x f G e tya tyb
    (HT: hastypepre  F ((x, tya) :: (f, Tarrow tya tyb) :: G) e tyb)
    (IH: cps_correct F ((x, tya) :: (f, Tarrow tya tyb) :: G) e tyb),
  cps_correct F G (Efix f x e) (Tarrow tya tyb).
Proof.
  exists 1; i; rrdes IH; s; s in UNQ; s in DJ.
  destruct (cps e _ _ _) eqn:E; destruct (ka _ _) eqn:K; s; s in UNQ; s in DJ.
  prepare; eapply pesim_dtsteps; [by eauto| | |by eauto|].
  - econs; [by apply mid_wdet; s; auto| |by econs]. 
    inst_pairs; eapply IRok; econs; eauto; econs.
  - rr; eauto 10.
  - i; rrdes XTR RLb; clarify; s in Ec; clarify; esplits.
    2: by right; split; eauto; instantiate (1:=wfo_nat i); econs; eauto
    ; do 2 econs; eauto; s; omega.
    2: by rr; esplits; eauto; inst; instantiate (1:=(_,Just _,Just _)); s; eauto.
    simpl cplus in HR; noemp.
    rewrite ->mid_csplit, src_csplit; eapply (@pesim_release _ _ cpsw); rr; eauto.

    assert (CL': forall a,
                  ~ fv_exp (Vfix f x (subst_exp [Evar x//x]>>[Evar f//f]>>denvb e)) a).
    { simpl fv_exp; i; des.
      eapply subst_exp_closed; eauto; i.
      destruct (LEM (x0 = x)) eqn:Ex.
      { subst; rewrite tmext_eq in FH2; s in FH2; subst; vauto. }
      rewrite tmext_neq in FH2; auto.
      destruct (LEM (x0 = f)) eqn:Ef.
      { subst; rewrite tmext_eq in FH2; subst; vauto. }
      rewrite tmext_neq in FH2; auto.
      eapply hastype_fv in FV; last first; eauto.
      s in FV; des; eauto. }

    assert (SIM: vclos cpsw s (g s) (tsubst_name tsub (Tarrow tya tyb))
                       (IVclo enva (IEfix (next n) (next (next n)) (next (next (next n))) e0))
                       (Vfix f x (subst_exp [Evar x//x]>>[Evar f//f]>>denvb e))).
    {
      clear KSIM.
      
      do 3 econs; [by econs|econs; [by econs|by eauto]|].
      apply GK; econs.
      + econs; [by econs; i; inv FH|]; rr; eauto.
      + depgen denvb; depgen enva.
        instantiate (1 := wfo_nat (i' + i)).
        pcofix CIH; i.
        apply pesim_gknow; i.
        invc STK; [|by invc INSTa RF VR].
        invc APPa APPb; invc RF RK VR; s in VAL; clarify.
        fold (snd (n0, e2)); rewrite <-E.
        rewrite <-subexp_tmext_l2; last first.
        * eauto.
        * eapply ms_gw_vclos_closed; eauto.
        * eauto using subclo_weakly.
        * apply pesim_gknow_; i.           
          { hexploit IH.
            - eauto.
            - instantiate (1:=s'); eauto.
            - instantiate (2:= (inr (next (next (next n))), ka0) ::
                 (inl (inr (next (next n))), va) ::
                 (inl (inr (next n)), IVclo env (IEfix (next n) (next (next n)) (next (next (next n)))
                                                 (snd (cps e (next (next (next (next n))))
                                                           [inr (next (next n))//x0]>>[inr (next n)//f0]>>ren
                                                           (fun mx y => (mx, IEret (next (next (next n))) y)))))) :: env).
              i; hexploit RSV; eauto; i; des.
              esplits; last first.
              + eapply vclos_mon; eauto.
                i; eapply (gk_mon GK0); eauto.
              + erewrite cpsw_ext; eauto.
              + eauto.
            - i.
              instantiate (1:= [vb // x0]>>
                 [Vfix f0 x0 (subst_exp [Evar x0 // x0]>>[Evar f0 // f0]>>denvb e) // f0]>>
                 denvb).
              instantiate (1:= [inr (next (next n))//x0]>>[inr (next n)//f0]>>ren).
              { apply list_assoc_cons_or in IN; des; subst.
                { rewrite !tmext_eq.
                  rewrite list_assoc_cons_neq; [|by unfold next; ii; clarify; omega].
                  rewrite list_assoc_cons_eq; eauto. }
                apply list_assoc_cons_or in IN0; des; subst.
                { rewrite tmext_neq; auto; rewrite tmext_eq.
                  rewrite tmext_neq; auto; rewrite tmext_eq.
                  rewrite list_assoc_cons_neq; [|by unfold next; ii; clarify; omega].
                  rewrite list_assoc_cons_neq; [|by unfold next; ii; clarify; omega].
                  rewrite list_assoc_cons_eq.
                  esplits; eauto.
                  
                  econs; econs; econs; s; eauto.
                  apply GK1; econs.
                  - econs.
                    + econs; ii; invc FH.
                    + econs; [by econs|]; eauto.
                  - i; apply gres_incl_id; right; eapply CIH; eauto using ext_trans.
                    fold (snd (n0, e2)); rewrite <-E; eauto.
                }
                rewrite tmext_neq; auto.
                rewrite tmext_neq; auto.
                rewrite tmext_neq; auto.
                rewrite tmext_neq; auto.
                rewrite list_assoc_cons_neq; last first.
                { unfold next; ii.
                  hexploit RDV; eauto; i; des. vauto.
                }
                rewrite list_assoc_cons_neq; last first.
                { unfold next; ii.
                  hexploit RDV; eauto; i; des.
                  clarify; rewrite <-H in *; eauto.
                }
                rewrite list_assoc_cons_neq; last first.
                { unfold next; ii.
                  hexploit RDV; eauto; i; des.
                  clarify; rewrite <-H in *; eauto.
                }
                hexploit RDV; eauto; i; des.
                esplits; eauto.
                eapply vclos_mon; eauto.
                i; eapply GK1; eauto.
              }
            - simpl map.
              instantiate (1 := fun mx y => (mx, IEret (next (next (next n))) y)).
              instantiate (1 := (next (next (next (next n))))).
              rewrite E.                      
              eapply uniq_appD in UNQ3; des; eauto.
            - rewrite E; clean; i.
              apply list_assoc_cons_or in LU; des; subst; eauto 10 using in_or_app.
              apply list_assoc_cons_or in LU0; des; subst; eauto 10 using in_or_app.
              apply list_assoc_cons_or in LU1; des; subst; eauto 10 using in_or_app.
            - rewrite <-tmext_lift; eapply weakclo_tmext; last first.
              { i; exfalso; eauto using ms_gw_vclos_closed. }
              rewrite <-tmext_lift; eapply weakclo_tmext; last first.
              { i; exfalso; eauto using ms_gw_vclos_closed. }
              eauto.
            - i; simpl map in IN.
              apply in_consD in IN; des; subst.
              { rewrite tmext_eq in FH; eauto using ms_gw_vclos_closed. }
              rewrite tmext_neq in FH; auto.
              apply in_consD in IN; des; subst.
              { rewrite tmext_eq in FH; eauto using ms_gw_vclos_closed. }
              rewrite tmext_neq in FH; eauto.
            - eapply cps_ksim_done.
              rewrite list_assoc_cons_eq; eauto.
            - eauto.
            - i; rewrite 2!tmext_lift.
              eapply paco10_mon; vauto.
          }
    }
    + unfold cps_ksim in KSIM.
      assert (LE: next (next (next (next n))) <= n0).
      { exploit cps_uniqvar; eauto; inst; [by s; i; clarify|].
        i; des; rrdes x0; eauto. }
      eapply paco10_mon; [eapply KSIM|by eauto].
      * eauto.
      * vauto.
      * eauto.
      * eauto.
      * eauto.
      * s; instantiate (1:=inr n); instantiate (1:=n0); rewrite K; eauto.
      * eauto.
      * i; eapply list_assoc_cons_fresh; eauto.
      * clean; rewrite K.
        eapply uniq_appD in UNQ3; des; eauto.
      * clean; rewrite K; i.
        apply list_assoc_cons_or in LU; des; subst; eauto 10 using in_or_app.
Qed.

Lemma cps_correct_app: forall F G ef e tya tyb
    (IHa: cps_correct F G ef (Tarrow tyb tya))
    (IHb: cps_correct F G e tyb),
  cps_correct F G (Eapp ef e) tya.
Proof.
  i; rrdes IHa IHb; eexists; i; s; s in UNQ.
  change (app_cont kb _) with (app_cont (SKappL (subst_exp denvb e)::kb) (subst_exp denvb ef)).
  instantiate (1 := _ + _); rewrite <-plus_assoc.
  eapply IHa; eauto; clear IHa; ii; subst; s.
  change (app_cont kb _) with (app_cont (SKappR vb::kb) (subst_exp denvb e)).
  instantiate (1 := _ + _); rewrite <-plus_assoc.
  eapply pesim_gknow; i.
  eapply IHb; eauto.
  - clear IHb.
    i; hexploit RSV; eauto; i; des.
    esplits; last first.
    + eapply vclos_mon; eauto.
      i; eapply (gk_mon GK0); eauto.
    + erewrite cpsw_ext; eauto.
    + eauto.                          
  - i; hexploit RDV; eauto; i; des.
    esplits; last first.
    + eapply vclos_mon; eauto.
      i; eapply (gk_mon GK0); eauto.
    + eauto.
  - ii; subst; s; s in UNQ1; s in DJ1; destruct (ka _ _) eqn:K; s; s in UNQ1; s in DJ1.
    invc RV REL VR; rrdes CB; invc CA SVQ; s in H0; clarify.
    prepare; eapply pesim_dtsteps; [by eauto| | |by eauto|].
    + eapply dtsteps_some; [by eapply mid_wdet; s; eauto| |].
      * inst_pairs; eapply IRok; econs; s; eauto.
      * 
        eapply dtsteps_none.
    + rr; esplits; [|eauto| |]; eauto.
    + i; rrdes RLb; invc XTR.
      s in Ec; clarify; noemp; esplits.
      2: by right; split; eauto; instantiate (1 := 1)
          ; instantiate (1 := wfo_nat i); econs; eauto 
          ; eapply wfo_nat_lt_mon; omega.
      
      2: by rr; esplits; eauto; inst; instantiate (1:=(_,Just _,Just _)); s.
      
      pfold; econs; i; esplits; auto.
      invc CNF; des; rrdes MA MB.
      invc XTRa XTRb; noemp; s in Ec; s in Ec0; clarify.
      apply esim_step; auto.
      
      i; invc ST; last first.
      * { exfalso; eapply STK; econs.
        - rewrite list_assoc_cons_neq; [eauto|ii; vauto].
        - s in LE2.
          rewrite list_assoc_cons_neq; [eauto|ii; vauto].
        - rewrite list_assoc_cons_eq; eauto.
        }
      * invc STEP.
        { esplits; first (by eauto); last (by auto); [right; econs; eauto|]
          ; last first; last 1 first.
          - inst_pairs; apply Rok; apply red_ctx; auto.
          - econs; first 2 last.
            + rr; esplits; eauto.
              instantiate (2:=(pmemp,Just _)); instantiate (1:=(_,Nothing)).
              s; noemp; auto.
            + rr; esplits; eauto.
              instantiate (2:=(pmemp,Nothing,Just _)).
              instantiate (1:=(_,Just _,Nothing)).
              s; noemp; auto.
            + auto.
            + split; rr; eauto.
          - 
            eapply call_app; eauto; last first; first 1 last.
            + econs.
              * econs.
              * econs; [|by eauto]; inst; rewrite <-H; econs; eauto.
              * eapply gk_mon; eauto.
            + econs; eauto; rewrite <-H; eauto.
            + rr.
              assert (XXX: Some (IVclo xvs'0 (IEfix bf0 bx0 bk0 e')) = Some (IVclo env (IEfix bf bx bk e1))).
              {
                rewrite list_assoc_cons_neq in GET; [|by ii; vauto].
                rewrite <-GET; eapply LE2, LEx.
              }
              clarify.
              assert (XXX: Some va0 = Some v).
              {
                destruct x0; symmetry; [by s in GETV; rewrite <-GETV; eapply LEx0|].
                rewrite list_assoc_cons_neq in GETV; eauto.
                rewrite <-GETV; apply LEx0.
              }
              clarify.
            + invc GETK; des_ifH H1; clarify; econs; eauto.
              i; eapply gres_incl_id; left.
              
              invc RETa; clear HR; clear dependent h.
              prepare; eapply pesim_dtsteps; [by eauto| | |by eauto|].
              * { eapply dtsteps_some.
                - apply mid_wdet; s; auto.
                - instantiate (1:=(_,_)); eapply IRok; econs; eauto.
                - eapply dtsteps_none.
                }
              * inst_pairs; rr; eauto 10.
              * { i; esplits; first 1 last.
                  - right; split; eauto.
                    instantiate (1 := wfo_nat (S _)).
                    instantiate (2 := wfo_nat _).
                    econs; eauto.
                    eapply wfo_nat_lt_mon; eauto.
                  - eauto.
                  - inv RETb; s; noemp.
                    simpl cplus in HR; noemp.
                    rewrite ->mid_csplit, src_csplit; eapply pesim_release; rr; eauto.
                    fold (snd (n2, e0)); rewrite <-K.
                    eapply KSIM; (safe eauto using ext_trans, extpub_trans)
                    ; first (by eauto); first 1 last; eauto.
                    + i; eapply LE0, LE2 in IN.
                      eapply list_assoc_cons_fresh; eauto.
                    + clean; rewrite K.
                      eapply uniq_appD in UNQ3; des; eauto.
                    + clean; rewrite K; i.
                      apply list_assoc_cons_or in LU; des; subst; eauto 10 using in_or_app.
                }
        }
Qed.

Lemma cps_correct_inst: forall F G e ty tyi
    (IH: cps_correct F G e (Tforall ty))
    (LC: lc_typ 0 tyi),
  cps_correct F G (Einst e) (tsubst_bvar 0 tyi ty).
Proof.
  i; rrdes IH; eexists; i; s; s in UNQ.
  change (app_cont kb _) with (app_cont (SKinst::kb) (subst_exp denvb e)).
  instantiate (1 := _ + _); rewrite <-plus_assoc.
  eapply IH; eauto; clear IH; ii; subst; s; s in UNQ0; s in DJ0.
  destruct (ka _ _) eqn:K; s; s in UNQ0; s in DJ0.
  invc RV REL VR; rrdes CB; invc CA SVQ; s in H0; clarify.
  prepare; eapply pesim_dtsteps; [by eauto| | |by eauto|].
    + eapply dtsteps_some; [by eapply mid_wdet; s; eauto| |].
      * instantiate (1:=(_,_)); eapply IRok; econs; s; eauto.
      * 
        eapply dtsteps_none.
    + rr; esplits; [|eauto| |]; eauto.
    + i; rrdes RLb; invc XTR.
      s in Ec; clarify; noemp; esplits.
      2: by right; split; eauto; instantiate (1 := 1)
          ; instantiate (1 := wfo_nat i); econs; eauto 
          ; eapply wfo_nat_lt_mon; omega.
      2: by rr; esplits; eauto; inst; instantiate (1:=(_,Just _,Just _)); s.
      
      pfold; econs; i; esplits; auto.
      invc CNF; des; rrdes MA MB.
      invc XTRa XTRb; noemp; s in Ec; s in Ec0; clarify.
      apply esim_step; auto.
      i; invc ST; last first.
      * { exfalso; eapply STK; econs.
        - rewrite list_assoc_cons_neq; [eauto|ii; vauto].
        - rewrite list_assoc_cons_eq; eauto.
        }
      * invc STEP.
        { esplits; first (by eauto); last (by auto); [right; econs; eauto|]
          ; last first; last 1 first.
          - inst_pairs; apply Rok; apply red_ctx; auto.
          - econs; first 2 last.
            + rr; esplits; eauto.
              instantiate (2:=(pmemp,Just _)); instantiate (1:=(_,Nothing)).
              s; noemp; auto.
            + rr; esplits; eauto.
              instantiate (2:=(pmemp,Nothing,Just _)).
              instantiate (1:=(_,Just _,Nothing)).
              s; noemp; auto.
            + auto.
            + split; rr; eauto.
          - 
            eapply call_inst; last first; first 1 last.
            + instantiate (1:=tsubst_name tsub tyi).
              eapply lc_tsubname; eauto.
            + econs.
              * econs.
              * econs; [|by eauto]; inst; rewrite <-H; econs; eauto.
              * eapply gk_mon; eauto.
            + econs; eauto; rewrite <-H; eauto.
            + rr.
              assert (XXX: Some (IVclo xvs'0 (IEgen bk0 e')) = Some (IVclo env (IEgen bk e1))).
              {
                rewrite list_assoc_cons_neq in GET; [|by ii; vauto].
                rewrite <-GET; eapply LEx.
              }
              clarify.
            + invc GETK; des_ifH H1; clarify; econs; eauto.
              i; eapply gres_incl_id; left.
              
              invc RETa; clear HR; clear dependent h.
              prepare; eapply pesim_dtsteps; [by eauto| | |by eauto|].
              * { eapply dtsteps_some.
                - apply mid_wdet; s; auto.
                - instantiate (1:=(_,_)); eapply IRok; econs; eauto.
                - eapply dtsteps_none.
                }
              * inst_pairs; rr; eauto 10.
              * { i; esplits; first 1 last.
                  - right; split; eauto.
                    instantiate (1 := wfo_nat (S _)).
                    instantiate (2 := wfo_nat _).
                    econs; eauto. 
                    apply wfo_nat_lt_mon; eauto.
                  - eauto.
                  - inv RETb; s; noemp.
                    simpl cplus in HR; noemp.
                    rewrite ->mid_csplit, src_csplit; eapply pesim_release; rr; eauto.
                    fold (snd (n1, e0)); rewrite <-K.
                    eapply KSIM; (safe eauto using ext_trans, extpub_trans)
                    ; first (by rewrite tsub_name_bvar; eauto); first 1 last; eauto.
                    + i; eapply LE0 in IN.
                      eapply list_assoc_cons_fresh; eauto.
                    + clean; rewrite K.
                      eapply uniq_appD in UNQ2; des; eauto.
                    + clean; rewrite K; i.
                      apply list_assoc_cons_or in LU; des; subst; eauto 10 using in_or_app.
                }
        }
Qed.

Lemma cps_correct_roll: forall F G e ty
    (IH: cps_correct F G e (tsubst_bvar 0 (Tmu ty) ty)),
  cps_correct F G (Eroll e) (Tmu ty).
Proof.
  i; rrdes IH; eexists; i; s; s in UNQ.
  change (app_cont kb _) with (app_cont (SKroll::kb) (subst_exp denvb e)).
  eapply IH; clear IH; eauto; ii; subst; s; s in UNQ0; s in DJ0.
  destruct (ka _ _) eqn:K; s; s in UNQ0; s in DJ0; eapply KSIM; eauto.
  - rewrite tsub_name_bvar in RV; eauto.
    econs; econs; s; eauto; rr; s; split; eauto using ms_gw_vclos_closed.
  - fold (snd (n1, e0)); rewrite <-K; eauto.
  - clean; rewrite K; eauto.
  - clean; rewrite K; eauto.
Qed.

Lemma cps_correct_unroll: forall F G e ty
    (IH: cps_correct F G e (Tmu ty)),
  cps_correct F G (Eunroll e) (tsubst_bvar 0 (Tmu ty) ty).
Proof.
  i; rrdes IH; eexists; i; s; s in UNQ.
  change (app_cont kb _) with (app_cont (SKunroll::kb) (subst_exp denvb e)).
  instantiate (1 := _ + _); rewrite <-plus_assoc.
  eapply IH; clear IH; eauto; ii; subst; s; s in UNQ0; s in DJ0.
  destruct (ka _ _) eqn:K; s; s in UNQ0; s in DJ0.
  inv RV REL; [|by destruct ty0 as [[]]; vauto]; rrdes CB; invc CA SVQ.
  prepare; eapply pesim_stepb_obsolete; i.
  - inv XTR; rrdes RL; s in Ec; noemp; clarify; esplits.
    + econs; inst_pairs; econs; eauto.
    + rr; esplits; eauto; inst; inst_pairs; s; noemp; eauto.
  - rewrite ->mid_csplit, src_csplit; apply pesim_release; rr; s; noemp; eauto.
    eapply KSIM; eauto.
    + rewrite tsub_name_bvar; eauto.
    + fold (snd (n1, e0)); rewrite <-K; s; noemp; eauto.
    + clean; rewrite K; eauto.
    + i; clean; rewrite K in IN; eauto.
Grab Existential Variables.
  exact 0.
Qed.

Lemma cps_correct_pack: forall F G e tyi ty
    (IH: cps_correct F G e (tsubst_bvar 0 tyi ty))
    (CLO: lc_typ 0 tyi),
  cps_correct F G (Epack e) (Texists ty).
Proof.
  i; rrdes IH; eexists; i; s; s in UNQ.
  change (app_cont kb _) with (app_cont (SKpack::kb) (subst_exp denvb e)).
  eapply IH; clear IH; eauto; ii; subst; s; s in UNQ0; s in DJ0.
  destruct (ka _ _) eqn:K; s; s in UNQ0; s in DJ0.
  eapply KSIM; eauto.
  - simpl tsubst_name; econs; econs; last first. 
    + rewrite <-tsub_name_bvar; eauto.
    + rr; s; split; eauto using ms_gw_vclos_closed.
    + s; eauto.
    + eapply lc_tsubname; eauto.
  - fold (snd (n1, e0)); rewrite <-K; eauto.
  - clean; rewrite K; eauto.
  - clean; rewrite K; eauto.
Qed.

Lemma cps_correct_unpack: forall F G x aa el er tyl tyr
    (IHl: cps_correct F G el (Texists tyl))
    (FTF: ~ ftn_typ_l (map (@snd _ _) F) aa)
    (FTG: ~ ftn_typ_l (map (@snd _ _) G) aa)
    (FTl: ~ ftn_typ tyl aa)
    (FTr: ~ ftn_typ tyr aa)
    (IHr: cps_correct F ((x, tsubst_bvar 0 (Tname aa) tyl) :: G) er tyr),
  cps_correct F G (Eunpack x el er) tyr.
Proof.
  i; rrdes IHl IHr; eexists; i; s; s in UNQ.
  replace (app_cont kb _) with
    (app_cont (SKunpack x (subst_exp [Evar x//x]>>denvb er)::kb)
              (subst_exp denvb el))
  ; [|by eauto].                
  instantiate (1 := _ + _); rewrite <-plus_assoc.
  eapply IHl; clear IHl; eauto; ii; subst; s; s in UNQ0; s in DJ0.
  destruct (cps er _ _ _) eqn:Er; s; s in UNQ0; s in DJ0.
  inv RV REL; [|by destruct ty as [[]]; vauto]; rrdes CB; invc CA SVQ.
  prepare; eapply pesim_stepb_obsolete; i.
  - inv XTR; rrdes RL; s in Ec; noemp; clarify; esplits.
    + econs; inst_pairs; econs; eauto.
    + rr; esplits; eauto; inst; inst_pairs; s; noemp; eauto.
  - rewrite ->mid_csplit, src_csplit; apply pesim_release; try by rr; s; noemp; eauto.
    s; noemp; fold (snd (n1, e)); rewrite <-Er.
    rewrite <-subexp_tmext_l, tmext_lift.
    eapply pesim_gknow; i.
    eapply paco10_mon; [eapply (IHr [tyi//aa]>>tsub)|by eauto]; safe eauto.
    + erewrite cpsw_ext; eauto.
      i; hexploit RSV; eauto; i; des; esplits; eauto.
      erewrite tsubname_agree.
      * eapply vclos_mon; eauto; i; eapply (gk_mon GK0); eauto.
      * i; clear -FTF IN IN0.
        assert (x1 <> aa).
        { ii; subst; induction F; vauto.
          destruct a; apply list_assoc_cons_or in IN; des; subst
          ; [apply FTF|eapply IHF]; ss; eauto.
        }
        rewrite tmext_neq; eauto.
    + i; apply list_assoc_cons_or in IN; des; subst.
      { rewrite !tmext_eq; esplits; eauto.
        rewrite tsub_name_bvar.
        - simpl tsubst_bvar.
          erewrite tmext_eq, tsubname_agree; eauto; i.
          assert (x1 <> aa) by (ii; subst; auto).
          rewrite tmext_neq; eauto.
        - i; destruct (LEM (x1 = aa)); subst.
          
          + rewrite tmext_eq; eauto.
          + rewrite tmext_neq; eauto.
      }
      rewrite !tmext_neq; eauto.
      hexploit RDV; eauto; i; des.
      esplits; eauto.
      erewrite tsubname_agree.
      * eapply vclos_mon; eauto; i; eapply (gk_mon GK0); eauto.
      * ii; clear -FTG IN1 IN0.
        assert (x2 <> aa).
        { ii; subst; induction G; vauto.
          destruct a; apply list_assoc_cons_or in IN0; des; subst
          ; [apply FTG|eapply IHG]; ss; eauto.
        }
        rewrite tmext_neq; eauto.
    + rewrite Er; eauto.
    + rewrite Er; eauto.
    + rewrite <-tmext_lift; eapply weakclo_tmext; last first.
      { i; exfalso; eauto using ms_gw_vclos_closed. }
      eauto.
    + i; apply in_consD in IN; des; subst.
      { rewrite tmext_eq in FH; eauto using ms_gw_vclos_closed. }
      rewrite tmext_neq in FH; eauto.
    + ii; eapply paco10_mon; [eapply KSIM|by eauto]; eauto using ext_trans.
      erewrite tsubname_agree; eauto.
      i; assert (x2 <> aa) by (ii; subst; auto).
      rewrite tmext_neq; eauto.
    + ii; destruct (LEM (x1 = aa)); subst.
      * rewrite tmext_eq; eauto.
      * rewrite tmext_neq; eauto.
    + eauto using subclo_weakly.
Grab Existential Variables.
  exact 0.
Qed.

Lemma cps_correct_main: forall F G e ty
    (HT: hastype F G e ty),
  cps_correct F G e ty.
Proof.
  i; invc HT; clear LCE0 LCT LCE; ginduction HT0; i.
  - apply cps_correct_input.
  - apply cps_correct_output; auto.
  - apply cps_correct_var; auto.
  - apply cps_correct_lbl; auto.
  - apply cps_correct_unit.
  - apply cps_correct_nat.
  - apply cps_correct_bop; auto.
  - apply cps_correct_ifnz; auto.
  - eapply cps_correct_pair; auto.
  - eapply cps_correct_fst; eauto.
  - eapply cps_correct_snd; eauto.
  - eapply cps_correct_inl; eauto.
  - eapply cps_correct_inr; eauto.
  - eapply cps_correct_case; eauto.
  - eapply cps_correct_roll; auto.
  - eapply cps_correct_unroll; auto.
  - eapply cps_correct_fix; eauto.
  - eapply cps_correct_app; eauto.
  - eapply cps_correct_gen; eauto.
  - eapply cps_correct_inst; eauto.
  - eapply cps_correct_pack; eauto.
  - eapply cps_correct_unpack; ss; eauto.
  - eapply cps_correct_new; auto.
  - eapply cps_correct_read; auto.
  - eapply cps_correct_write; eauto.
  - eapply cps_correct_refeq; eauto.
Qed.


Lemma lookup3: forall K (M: EqDec K) X Y Z xsl xsr ysl ysr zsl zsr P (x: X) (y: Y) (z: Z) k
    (Eyx: map fst ysl = map fst xsl)
    (Ezx: map fst zsl = map fst xsl)
    (INx: lookup k (xsl ++ xsr) = Some x)
    (INy: lookup k (ysl ++ ysr) = Some y)
    (INz: lookup k (zsl ++ zsr) = Some z)
    (Pl: forall k x y z
             (INx: lookup k xsl = Some x)
             (INy: lookup k ysl = Some y)
             (INz: lookup k zsl = Some z),
           P x y z : Prop)
    (Pr: forall k x y z
             (INx: lookup k xsr = Some x)
             (INy: lookup k ysr = Some y)
             (INz: lookup k zsr = Some z),
           P x y z),
    P x y z.
Proof.
  i.  
  apply list_assoc_app_or in INx; apply list_assoc_app_or in INy.
  apply list_assoc_app_or in INz; des; eauto
  ; exfalso; eauto using list_assoc_eqdom_false.
  rewrite <-Eyx in Ezx; exfalso; eauto using list_assoc_eqdom_false.
Qed.





Lemma cps_tl_correct: forall FI suba subb envf refdb (fty: flextyp) (v: src_val) n
    (SUBa: map fst suba = map fst FI)
    (SUBb: map fst subb = map fst FI)
    (TE: is_thunkenv (map_pair id flext_of FI))
    (TT: is_thunktyp fty)
    (HT: hastype (map_pair id flext_of FI) [] v fty),
  @tlsim _ _ ms_gw (lwemp ms_gw) FI suba subb (tt, subb ++ envf, refdb, tt)
        (IVclo (map_pair (inl<*>inl) id suba) (snd (tlcps n v))) v fty.
Proof.
  ii; inv HT; destruct v; (try by ss; destruct e_of; vauto)
  ; inv HT0; try (by destruct fty; ss; destruct flext_of; vauto).
  - destruct e_of; vauto.
    unfold tlcps; destruct (cps _ _ _ _) eqn:E; unfold snd.
    assert (fty = FTarrow tya tyb).
    { apply flext_of_inj; destruct fty; s; destruct flext_of; vauto. }
    subst; clear H; s in H1; clarify; econs; first by econs.
    { rr; s; split; eauto.
      ii; des; eapply hastype_fv in FH1; eauto.
      s in FH1; des; vauto.
    }
    hexploit cps_correct_main
    ; [by invc LCT; econs; eauto; ii; s in IN; des; vauto|].
    intro SIM; rrdes SIM.

    eapply gk_lsim; eauto.
    econs.
    { econs.
      - econs; i; inv FH.
      - econs; [by econs|].
        s; ii; des; eapply hastype_fv in FH1; eauto.
        s in FH1; des; vauto.
    }

    { instantiate (1 := wfo_nat (i' + 1)).
      pcofix CIH; i.
      apply pesim_gknow; i; apply pesim_gknow_; i.
      invc STK; [|by invc INSTa RF VR].
      invc APPa APPb; invc RF RK VR; s in VAL; clarify.
      fold (snd (n0, e)); rewrite <-E.
      rewrite !(tmext_lift e_of).
      eapply paco10_mon; [eapply SIM|by eauto]; safe eauto.
      - i.
        assert (exists fty, lookup l FI = Some fty /\ ty = fty).
        { apply list_assoc_map_pairD; auto. }
        des.
        assert (exists va', lookup l suba = Some va').
        { apply eq_sym in SUBa; eapply list_assoc_eqdom; eauto. }
        des.
        assert (exists vb', lookup l subb = Some vb').
        { apply eq_sym in SUBb; eapply list_assoc_eqdom; eauto. }
        des.
        hexploit RS; eauto; i.
        rewrite list_assoc_cons_neq; vauto.
        rewrite list_assoc_cons_neq; vauto.
        rewrite list_assoc_cons_neq; vauto.
        esplits; last first.
        + instantiate (3:=tid).
          rewrite tsubst_name_tid.
          econs; eapply vclgen_thunk.
          eapply vclos_thunk_mon; eauto.
          i; eapply GK0; eauto.
        + erewrite cpsw_ext; eauto.
          erewrite cpsw_ext; eauto.
          rewrite list_assoc_app H3; eauto.
        + clear -H1; induction suba; auto.
          destruct a; unfold map_pair; simpl map.
          unfold list_assoc; des_if.
          * clarify; rewrite list_assoc_cons_eq in H1; clarify.
          * rewrite list_assoc_cons_neq in H1; [|by ii; vauto].
            eauto.
      - i; s in IN; des_ifH IN; clarify.
        + rewrite !tmext_eq.
          rewrite list_assoc_cons_neq; [|by vauto].
          rewrite list_assoc_cons_eq.
          esplits; eauto.
          rewrite tsubst_name_tid; eauto.
        + des_ifH IN; vauto.    
          rewrite tmext_neq; [|by auto].
          rewrite tmext_eq.
          rewrite tmext_neq; [|by auto].
          rewrite tmext_eq.
          rewrite list_assoc_cons_neq; [|by vauto].
          rewrite list_assoc_cons_neq; [|by unfold next; ii; vauto; omega].
          rewrite list_assoc_cons_eq.
          esplits; eauto.
          rewrite tsubst_name_tid.
          econs; eapply vclgen_thunk.
          econs; [by econs| |].
          { econs; [by rr; s; econs|].
            ii; eapply hastype_fv in FH; vauto; eauto. }
          eapply (gk_lsim GK1).           
          econs.
          { econs.
            - clear; econs; i; invc FH.
            - econs; [by econs|].
              ii; eapply hastype_fv in FH; vauto; eauto. }
          rewrite E.
          i; eapply gres_incl_id. right.
          eapply CIH; eauto using ext_trans.
      - rewrite E.
        hexploit cps_uniqvar; eauto; inst; [by i; s in E0; clarify|]; i; des.
        eapply (uniq_mapD _ varid); eauto.
      - rewrite E; clean; i.
        hexploit cps_uniqvar; eauto; inst; [by i; s in E0; clarify|]; i; des.
        apply list_assoc_cons_or in LU; des; subst.
        { apply in_split in IN; des.
          rewrite IN map_app in H0; ss; rrdes H0.
          hexploit H1; [by eapply in_or_app; right; left; eauto|].
          unfold next; omega. }
        apply list_assoc_cons_or in LU0; des; subst.
        { apply in_split in IN; des.
          rewrite IN map_app in H0; ss; rrdes H0.
          hexploit H1; [by eapply in_or_app; right; left; eauto|].
          unfold next; omega. }
        apply list_assoc_cons_or in LU1; des; subst.
        { apply in_split in IN; des.
          rewrite IN map_app in H0; ss; rrdes H0.
          hexploit H1; [by eapply in_or_app; right; left; eauto|].
          unfold next; omega. }
        assert (exists l, y = inl (inl l)).
        {
          clear -LU2; induction suba; vauto.
          destruct a; simpl map_pair in LU2.
          apply list_assoc_cons_or in LU2; des; subst; eauto.
        }
        des; subst; eauto using bvars_no_labels.
      - rewrite <-tmext_lift; eapply weakclo_tmext; last first.
        { i; exfalso; eauto using ms_gw_vclos_closed. }
        rewrite <-tmext_lift; eapply weakclo_tmext; eauto.
        { i; exfalso; eapply (@hastype_fv _ []) in H; vauto. }
      - i; apply in_consD in IN; des; subst.
        { rewrite tmext_eq in FH; eauto using ms_gw_vclos_closed. }
        rewrite tmext_neq in FH; auto.
        apply in_consD in IN; des; subst.
        { rewrite tmext_eq in FH; eauto using ms_gw_vclos_closed.
          eapply (@hastype_fv _ []) in FH; vauto. }
        rewrite tmext_neq in FH; eauto.
      - rewrite tsubst_name_tid; eapply cps_ksim_done; eauto.
      - eauto.
    }
  - destruct e_of; vauto.
    unfold tlcps.
    destruct (cps _ _ _ _) eqn:E. unfold snd.
    assert (fty = FTforall ty).
    { apply flext_of_inj; destruct fty; s; destruct flext_of; vauto. }
    subst; clear H; s in H1; clarify; econs; first by econs.
    { rr; s; split; eauto.
      ii; des; eapply hastype_fv in HT0; eauto.
      s; eauto.
    }
    inv HT0.
    hexploit cps_correct_main; [by econs; eauto; eapply tsubbvar_lctyp; eauto|].
    intro SIM; rrdes SIM.

    eapply gk_lsim; eauto.
    econs.
    { econs.
      - econs; i; invc FH.
        hexploit (cps_good e (next 0) inr (fun mx' y => (mx', IEret 0 y))); eauto.
        + ii; vauto.
        + rewrite E; s; ii; vauto.
      - econs.
        + s; eauto; ii.
        + ii; eapply hastype_fv in FH; eauto; vauto.
    }
    { instantiate (1 := wfo_nat (i' + 1)).
      pcofix CIH; i.
      apply pesim_gknow; i.
      apply pesim_gknow_; i.
      invc STK; [by invc APPa RF VR|].
      invc INSTa INSTb; invc RF RK VR; s in VAL; clarify.
      fold (snd (n0, e1)); rewrite <-E. s in H3. clarify.
      replace (app_cont kb e2) with (app_cont kb (subst_exp vid e2)); last first.
      rewrite subexp_vid; auto.
      eapply paco10_mon; [eapply (SIM [tyi//aa0]>>tid)|by eauto]; safe eauto.
      - i.
        assert (exists fty, lookup l FI = Some fty /\ ty = fty).
        { apply list_assoc_map_pairD; auto. }
        des.
        assert (exists va', lookup l suba = Some va').
        { apply eq_sym in SUBa; eapply list_assoc_eqdom; eauto. }
        des.
        assert (exists vb', lookup l subb = Some vb').
        { apply eq_sym in SUBb; eapply list_assoc_eqdom; eauto. }
        des.
        hexploit RS; eauto; i.
        rewrite list_assoc_cons_neq; vauto.
        esplits; last first.
        + erewrite tsubname_agree.
          * instantiate (3:=tid).
            rewrite tsubst_name_tid.
            econs; eapply vclgen_thunk.
            eapply vclos_thunk_mon; eauto.
            i; eapply GK0; eauto.
          * i; clear -H7 IN H IN0.
            assert (x <> aa0).
            { ii; subst; induction FI; vauto.
              destruct a; apply list_assoc_cons_or in H; des; subst
              ; [apply H7|eapply IHFI]; ss; i; des; eauto using IN0.
              des_ifH IN; vauto.
            }
            rewrite tmext_neq; eauto.
        + erewrite cpsw_ext; eauto.
          erewrite cpsw_ext; eauto.
          rewrite list_assoc_app H6; eauto.
        + clear -H3; induction suba; auto.
          destruct a; unfold map_pair; simpl map.
          unfold list_assoc; des_if.
          * clarify; rewrite list_assoc_cons_eq in H3; clarify.
          * rewrite list_assoc_cons_neq in H3; [|by ii; vauto].
            eauto.
      - i; vauto.
      - rewrite E.
        hexploit cps_uniqvar; eauto; inst; [by i; s in E0; clarify|]; i; des.
        eapply (uniq_mapD _ varid); eauto.
      - rewrite E; clean; i.
        hexploit cps_uniqvar; eauto; inst; [by i; s in E0; clarify|]; i; des.
        apply list_assoc_cons_or in LU; des; subst.
        { apply in_split in IN; des.
          rewrite IN map_app in H2; ss; rrdes H2.
          hexploit H3; [by eapply in_or_app; right; left; eauto|].
          unfold next; omega. }
        assert (exists l, y = inl (inl l)).
        {
          clear -LU0; induction suba; vauto.
          destruct a; simpl map_pair in LU0.
          apply list_assoc_cons_or in LU0; des; subst; eauto.
        }
        des; subst; eauto using bvars_no_labels.
      - rewrite tsub_name_bvar.
        + simpl tsubst_bvar.
          erewrite tmext_eq, tsubname_agree; eauto; i.
          erewrite tsubname_agree.
          * instantiate (1:=tid); rewrite tsubst_name_tid.
            eapply cps_ksim_done; eauto.
          * ii; rewrite tmext_neq; eauto.
            clear -H8 IN; ii; subst.
            apply H8; s; eauto.
        + i; destruct (LEM (x = aa0)); subst.
          * rewrite tmext_eq; auto.
          * rewrite tmext_neq; auto.
      - ii; destruct (LEM (x = aa0)); subst.
        + rewrite tmext_eq; auto.
        + rewrite tmext_neq; auto.
    }
Qed.

Theorem mcps_correct: forall FI FO FI' FO' m
    (WF: mhastype FI' m FO')
    (EFI: FI' = map_pair id flext_of FI)
    (EFO: FO' = map_pair id flext_of FO),
  msim ms_gw FI FO (mcps m) m.
Proof.
  exists (lwemp ms_gw); i; rrdes LMa LMb.
  exists (tt, subb ++ m ++ envf, refst_emp, tt).
  splits; s; vauto; i; [|by rrdes FH; vauto|].
  - eexists; ex @cemp; eexists; ex @cemp.
    splits; (noemp; auto); [by econs; econs|].
    rr; esplits; s; unfold mid_cemp, src_cemp; eauto using pmvalid_emp
    ; try (i; apply pmread_empD in ALC); vauto.
  - revert_until m; unfold mcps; generalize 0.
    ginduction m; i; s; inv WF MHT; destruct FO; vauto; [].
    rewrite cons_app in IN; destruct p.
    s in FLO; rrdes FLO; s in H3; clarify.
    hexploit cps_tl_correct; eauto.
    intro SIM; instantiate (3:= (l0,v)::m ++ envf) in SIM.
    instantiate (1:=n) in SIM.
    unfold id.                  
    destruct (tlcps _ _) eqn:E.
    unfold mid_vload; s.
    des_if.
    + 
      subst; esplits; rr; eauto.
      s in IN; des_ifs; vauto.
    + 
      set (v0 := IVclo (map (fun p => (inl (inl (fst p)), snd p)) suba) e).
      hexploit (IHm (FI ++ [(l0, f0)])).
      * eauto.
      * instantiate (1 := suba ++ [(l0, v0)]).
        rewrite !map_app; s; f_equal; eauto.
      * instantiate (1 := subb ++ [(l0, v)]).
        rewrite !map_app; s; f_equal; eauto.
      * rewrite (map_pair_app eqdec); econs; eauto.
        apply is_thunkenv_app; s; eauto.
      * s in IN; des_ifH IN; vauto.
      * i; des; esplits; eauto.
        { rr; rewrite list_assoc_cons_neq; eauto. }
        rewrite cons_app; rewrite -app_assoc in H1.
        ii; exploit H1; eauto.
        i; eapply (lookup3 _ _ _ _ _ _ _ _ _ SUBa SUBb); eauto.
        i; sng; eapply vclos_thunk_mon; [apply SIM| |]; eauto.
Grab Existential Variables.
  exact tt. exact tt. exact s.
Qed.






